# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import resources
from libopensesame import metadata, misc
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.tab_bar import TabBar
from qtpy import QtGui, QtWidgets, QtCore
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'tab_widget', category=u'core')

# Determines how many tabs are kept active and automatically switched back to
# when the current tab is closed.
MAX_TAB_STACK = 3


class TabWidget(BaseSubcomponent, QtWidgets.QTabWidget):

    """A custom tab widget with some extra functionality"""
    tab_removed = QtCore.Signal(int)
    tab_inserted = QtCore.Signal(int)

    def __init__(self, parent=None):
        r"""Constructor

        Parameters
        ----------
        parent, optional
            The parent QWidget
        """
        QtWidgets.QTabWidget.__init__(self, parent)
        self.setTabBar(TabBar(self))
        BaseSubcomponent.setup(self, parent)
        self.tabCloseRequested.connect(self.removeTab)
        self.currentChanged.connect(self.index_changed)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.MinimumExpanding
        )
        self.shortcut_switch_left = QtWidgets.QShortcut(
            QtGui.QKeySequence.PreviousChild,
            self.main_window,
            self.switch_prev,
            self.switch_prev
        )
        self.shortcut_switch_right = QtWidgets.QShortcut(
            QtGui.QKeySequence.NextChild,
            self.main_window,
            self.switch_next,
            self.switch_next
        )
        self._tab_stack = []

    def switch_prev(self):
        r"""Switches to the previous tab."""
        i = self.currentIndex() - 1
        if i < 0:
            i = self.count() - 1
        self.setCurrentIndex(i)

    def switch_next(self):
        r"""Switches to the next tab."""
        i = self.currentIndex() + 1
        if i >= self.count():
            i = 0
        self.setCurrentIndex(i)

    def _update_actions(self):

        count = self.count()
        index = self.currentIndex()
        self.main_window.ui.action_close_other_tabs.setVisible(count > 1)
        self.main_window.ui.action_close_all_tabs.setVisible(count > 0)
        self.main_window.ui.action_close_current_tab.setVisible(count > 0)

    def tabRemoved(self, i):
        r"""Overridden to emit the tab_removed signal when a tab has been
        removed.
        """
        QtWidgets.QTabWidget.tabRemoved(self, i)
        self.tab_removed.emit(i)
        self._update_actions()

    def tabInserted(self, i):
        r"""Overridden to emit the tab_removed signal when a tab has been
        removed.
        """
        QtWidgets.QTabWidget.tabInserted(self, i)
        self.tab_inserted.emit(i)
        self._update_actions()

    def add(self, widget, icon, name, switch=True):
        r"""Open a tab, or switches to it if the tab already exists.

        Parameters
        ----------
        widget
            A QWidget for the tab
        icon
            The name of an icon or a QIcon
        name
            A name for the tab
        switch : bool, optional
            Indicates whether the tab should be switched to.
        """
        # If a widget is currently open, then we push it to the stack so that
        # we can go back to it later when the newly added widget is closed.
        current_index = self.currentIndex()
        current_widget = self.widget(current_index)
        if current_widget is not None:
            self._tab_stack.append((
                current_widget,
                self.tabIcon(current_index),
                self.tabText(current_index))
            )
            self._tab_stack = self._tab_stack[-MAX_TAB_STACK:]
        index = self.indexOf(widget)
        if index < 0:
            index = self.addTab(
                widget,
                self.main_window.theme.qicon(icon),
                name
            )
        if switch:
            self.setCurrentIndex(index)

    def remove(self, widget):
        r"""Removes the tab with a given widget in it.

        Parameters
        ----------
        widget : QWidget
            The widget in the tab.
        """
        index = self.indexOf(widget)
        if index >= 0:
            self.removeTab(index)

    def set_icon(self, widget, icon):
        r"""Sets the icon for a tab with a specific widget.

        Parameters
        ----------
        widget : QWidget
            A widget.
        icon : unicode
            An icon name.
        """
        index = self.indexOf(widget)
        if index >= 0:
            self.setTabIcon(index, self.theme.qicon(icon))

    def close_all(self, dummy=None, avoid_empty=True):
        r"""Closes all tabs.

        Parameters
        ----------
        avoid_empty : bool, optional
            Indicates whether the general tab should be automatically opened if
            all tabs are closed.
        """
        while self.count():
            self.removeTab(0)
        if avoid_empty and not self.count():
            self.open_general()

    def close_current(self, dummy=None, avoid_empty=True):
        r"""Closes the current tab.

        Parameters
        ----------
        avoid_empty : bool, optional
            Indicates whether the general tab should be automatically opened if
            all tabs are closed.
        """
        self.removeTab(self.currentIndex())
        if not avoid_empty or self.count():
            return
        if self._tab_stack:
            self.add(*self._tab_stack.pop())
        else:
            self.open_general()

    def close_other(self):
        """Close all tabs except for the currently opened one"""
        while self.count() > 0 and self.currentIndex() != 0:
            self.removeTab(0)
        while self.count() > 1:
            self.removeTab(1)

    def get_index(self, tab_name):
        """
        Return the index of a specific tab

        Arguments:
        tab_name -- the tab_name of the widget

        Returns:
        The index of the tab or None if the tab wasn't found
        """
        for i in range(self.count()):
            w = self.widget(i)
            if (hasattr(w, u"tab_name") and w.tab_name == tab_name) or \
                    (hasattr(w, tab_name)):
                return i
        return None

    def get_item(self, item):
        """
        Return the index of a specific item tab

        Arguments:
        item -- the name of the item

        Returns:
        The index of the tab or None if the tab wasn't found
        """
        for i in range(self.count()):
            w = self.widget(i)
            if (hasattr(w, u"__edit_item__") and w.__edit_item__ == item):
                return i
        return None

    def current_item(self):
        """
        returns:
                desc:	The name of the currently visible item, or None if no item
                                is currently visible.
                type:	[str, NoneType]
        """
        w = self.currentWidget()
        if hasattr(w, u'__item__'):
            return w.__item__
        return None

    def get_widget(self, tab_name):
        """
        Return a specific tab

        Arguments:
        tab_name -- the tab_name of the widget

        Returns:
        A QWidget or None if the tab wasn't found
        """
        i = self.get_index(tab_name)
        if i is None:
            return None
        return self.widget(i)

    def open_browser(self, url):
        """
        Open a browser tab to browse local or remote HTML files

        Argument:
        url -- a url
        """
        from libqtopensesame.widgets import webbrowser
        browser = webbrowser.webbrowser(self.main_window)
        browser.load(url)
        self.add(browser, u"applications-internet", _(u'Help'))

    def open_help(self, item):
        r"""Opens a help tab for the specified item. Looks for a file called
        [item].html or [item].md in the resources folder.

        Parameters
        ----------
        item
            The item for which help should be displayed.
        """
        try:
            path = resources[f'{item}.md']
        except FileNotFoundError:
            try:
                path = resources[f'{item}.html']
            except FileNotFoundError:
                path = resources[f'help/missing.md']
        self.open_browser(path)

    def open_backend_settings(self):
        """Opens the backend settings"""
        if self.switch(u'__backend_settings__'):
            return
        from libqtopensesame.widgets.backend_settings import backend_settings
        self.add(backend_settings(self.main_window), u'backend',
                 _(u'Back-end settings'))

    def open_forum(self):
        """Open osdoc.cogsci.nl"""
        self.open_browser(u'http://forum.cogsci.nl')

    def open_general(self):
        """Opens the general tab"""
        if self.switch(u'__general_properties__'):
            return
        from libqtopensesame.widgets.general_properties import general_properties
        w = general_properties(self.main_window)
        self.add(w, u'experiment', _(u'General properties'))

    def open_general_script(self):
        """Opens the general script editor"""
        if self.switch(u'__general_script__'):
            return
        from libqtopensesame.widgets.general_script_editor import \
            general_script_editor
        self.add(general_script_editor(self.main_window), u'utilities-terminal',
                 _(u'General script editor'))

    def open_osdoc(self, url=u''):
        r"""Open a page on osdoc.cogsci.nl."""
        misc.open_url(
            u'http://osdoc.cogsci.nl/%s/%s' % (metadata.main_version, url)
        )

    def open_stdout_help(self):
        """Open the debug window help tab"""
        self.open_osdoc(u'manual/interface')

    def open_unused(self):
        """Opens the unused tab"""
        if self.switch(u'__unused__'):
            return
        from libqtopensesame.widgets.unused_widget import unused_widget
        w = unused_widget(self.main_window)
        self.add(w, u'unused', _(u'Unused items'))

    def open_preferences(self):
        """Open the preferences tab"""
        from libqtopensesame.widgets import preferences_widget
        if not self.switch(u"__preferences__"):
            self.add(preferences_widget.preferences_widget(self.main_window),
                     u"options", _(u"Preferences"))

    def switch(self, tab_name):
        """
        Switch to a specific tab

        Returns:
        True if the tab exists, False otherwise
        """
        i = self.get_index(tab_name)
        if i is None:
            return False
        self.setCurrentIndex(i)
        return True

    def toggle_onetabmode(self):
        """Toggles onetabmode"""
        cfg.onetabmode = self.main_window.ui.action_onetabmode.isChecked()
        if cfg.onetabmode:
            self.close_other()
            self.tabBar().setVisible(False)
        else:
            self.tabBar().setVisible(True)
        self.setTabsClosable(not cfg.onetabmode)
        self.main_window.ui.action_close_all_tabs.setEnabled(
            not cfg.onetabmode)
        self.main_window.ui.action_close_current_tab.setEnabled(
            not cfg.onetabmode)
        self.main_window.ui.action_close_other_tabs.setEnabled(
            not cfg.onetabmode)

    def index_changed(self, index):
        """
        Monitors tab index changes, closing other tabs if onetabmode is enabled

        Arguments:
        index -- the index of the new tab
        """
        self.currentChanged.disconnect()
        if cfg.onetabmode:
            self.close_other()
        w = self.currentWidget()
        if hasattr(w, u'on_activate'):
            w.on_activate()
        self.currentChanged.connect(self.index_changed)
        self._update_actions()

    def open_markdown(self, md, icon=u'dialog-information',
                      title=u'Attention!', url=None, tmpl=None):
        r"""Opens a Markdown-formatted tab.

        Parameters
        ----------
        md : str
            If it ends with `.md`, this is interpreted as a file name.
            Otherwise, it is interpreted as markdown text.
        icon : str, optional
            The name of the tab icon.
        title : str, optional
            The tab title.
        url : str, optional
            A url to identify the page. This only applies when the markdown is
            provided as text; otherwise the markdown filename is used.
        """
        from libqtopensesame.widgets.webbrowser import webbrowser
        wb = webbrowser(self.main_window)
        if md.endswith(u'.md'):
            wb.load(md, tmpl=tmpl)
        else:
            wb.load_markdown(md, url=url, tmpl=tmpl)
        self.main_window.tabwidget.add(wb, icon, title)

    def focus(self):
        r"""Properly give focus to the current widget (if any)."""
        item = self.current_item()
        if item is None:
            return
        self.experiment.items[item].set_focus()


# Alias for backwards compatibility
tab_widget = TabWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtCore
from libopensesame import metadata
from libqtopensesame.misc.config import cfg
from libqtopensesame.widgets.webbrowser import WebView
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.misc import template_info
_ = translation_context('analytics', category='extension')


HTML = '''<!DOCTYPE html>
<html>
<head>
<!-- Google tag (gtag.js) -->
<script async src="https://www.googletagmanager.com/gtag/js?id=G-TPL6TXE7KL">
</script>
<script>
  window.dataLayer = window.dataLayer || [];
  function gtag(){{dataLayer.push(arguments);}}
  gtag('js', new Date());

  gtag('config', 'G-TPL6TXE7KL');
</script>
<title>{mode} ({version})</title>
</head>
</body>DUMMY BODY</body>
</html>
'''


class Analytics(BaseExtension):
    """Sends a ping to Google Analytics when OpenSesame is started."""
    
    def event_startup(self):
        wv = WebView(self.main_window)
        wv.setHtml(
            HTML.format(mode=self.main_window.mode,
                        version=metadata.__version__),
            QtCore.QUrl(f'http://opensesame.app.cogsci.nl/{self.main_window.mode}/{metadata.__version__}'))
        wv.hide()
        if not cfg.analytics_show_notification:
            return
        self.extension_manager.fire(
            'notify',
            message=_('Anonymous usage data is collected. You can disable '
                      'this by disabling the <i>analytics</i> extension.'),
            category='info', timeout=10000, buttontext=_('Got it!'))
        cfg.analytics_show_notification = False



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.var_store import var_store
import warnings
from libopensesame.misc import snake_case
from libopensesame.exceptions import InvalidOpenSesameScript, InvalidValue, \
    IncompatibilityError
from libopensesame.oslogging import oslogger


class Item:
    """Abstract class that serves as the basis for all OpenSesame items.
    
    Parameters
    ----------
    name: str
        The name of the item
    experiment: Experiment
    string: str or None, optional
        A definition string
    """
    
    encoding = u'utf-8'
    var = None

    def __init__(self, name, experiment, string=None):
        if self.var is None:
            self.var = var_store(self, parent=experiment.var)
        self.name = name
        self.experiment = experiment
        self.debug = oslogger.debug_mode
        self.count = 0
        self._get_lock = None
        # Deduce item_type from class name. This takes into account that the
        # class name can be CamelCase and may prefixed by Qt (QtCamelCase),
        # whereas the item_type should always be snake_case without the qt 
        # prefix.
        prefix = self.experiment.item_prefix()
        self.item_type = self.__class__.__name__
        if self.item_type.lower().startswith(prefix.lower()):
            self.item_type = self.item_type[len(prefix):]
        self.item_type = snake_case(self.item_type)
        if not hasattr(self, u'description'):
            self.var.description = self.default_description
        else:
            self.var.description = self.description
        self.from_string(string)

    def __deepcopy__(self, memo):
        """Disable deep-copying of items"""
        return None

    @property
    def clock(self):
        return self.experiment._clock

    @property
    def log(self):
        return self.experiment._log

    @property
    def syntax(self):
        return self.experiment._syntax

    @property
    def python_workspace(self):
        return self.experiment._python_workspace

    @property
    def responses(self):
        return self.experiment._responses
    
    @property
    def plugin_manager(self):
        return self.experiment._plugin_manager

    @property
    def default_description(self):
        return u'Default description'

    def reset(self):
        r"""Resets all item variables to their default value."""
        pass

    def prepare(self):
        """Implements the prepare phase of the item."""
        self.experiment.var.set(u'count_%s' % self.name, self.count)
        self.count += 1

    def run(self):
        """Implements the run phase of the item."""
        pass

    def parse_variable(self, line):
        """Reads a single variable from a single definition line.

        Parameters
        ----------
        line
            A single definition line.

        Returns
        -------
        bool
            True on success, False on failure.
        """
        # It is a little ugly to call parse_comment() here, but otherwise
        # all from_string() derivatives need to be modified
        if self.parse_comment(line):
            return True
        l = self.syntax.split(line.strip())
        if len(l) == 0 or l[0] != u'set':
            return False
        if len(l) != 3:
            raise InvalidOpenSesameScript(
                'Error parsing variable definition', line=line)
        self.var.set(l[1], l[2])
        return True

    def parse_line(self, line):
        """Allows for arbitrary line parsing, for item-specific requirements.

        Parameters
        ----------
        line
            A single definition line.
        """
        pass

    def parse_comment(self, line):
        """Parses comments from a single definition line, indicated by 
        # // or '.

        Parameters
        ----------
        line
            A single definition line.

        Returns
        -------
        bool
            True on success, False on failure.
        """
        line = line.strip()
        if len(line) > 0 and line[0] == u'#':
            self.comments.append(line[1:])
            return True
        elif len(line) > 1 and line[0:2] == u'//':
            self.comments.append(line[2:])
            return True
        return False

    def variable_to_string(self, var):
        r"""Encodes a variable into a definition string.

        Parameters
        ----------
        var : str
            The name of the variable to encode.

        Returns
        -------
        str
            A definition string.
        """
        val = safe_decode(self.var.get(var, _eval=False))
        # Multiline variables are stored as a block
        if u'\n' in val or u'"' in val:
            s = u'__%s__\n' % var
            val = val.replace(u'__end__', u'\\__end__')
            for l in val.split(u'\n'):
                s += '\t%s\n' % l
            while s[-1] in (u'\t', u'\n'):
                s = s[:-1]
            s += u'\n\t__end__\n'
            return s
        # Regular variables
        return self.syntax.create_cmd(u'set', arglist=[var, val]) + u'\n'

    def from_string(self, string):
        r"""Parses the item from a definition string.

        Parameters
        ----------
        string : str, NoneType
            A definition string, or None to reset the item.
        """
        self._script = string
        textblock_var = None
        self.var.clear()
        self.reset()
        self.comments = []
        if string is None:
            return
        for line in string.split(u'\n'):
            line_stripped = line.strip()
            # The end of a textblock
            if line_stripped == u'__end__':
                if textblock_var is None:
                    raise InvalidOpenSesameScript(
                        'It appears that a textblock has been closed without '
                        'being opened')
                self.var.set(textblock_var,
                             textblock_val.replace(u'\\__end__', u'__end__'))
                textblock_var = None
            # The beginning of a textblock. A new textblock is only started when
            # a textblock is not already ongoing, and only if the textblock
            # start is of the format __VARNAME__
            elif line_stripped[:2] == u'__' and line_stripped[-2:] == u'__' \
                    and textblock_var is None:
                textblock_var = line_stripped[2:-2]
                if textblock_var != u'':
                    textblock_val = u''
                else:
                    textblock_var = None
                # We cannot just strip the multiline code, because that may mess
                # up indentation. So we have to detect if the string is indented
                # based on the opening __varname__ line.
                strip_tab = line[0] == u'\t'
            # Collect the contents of a textblock
            elif textblock_var is not None:
                if strip_tab:
                    textblock_val += line[1:] + u'\n'
                else:
                    textblock_val += line + u'\n'
            # Parse regular variables
            elif not self.parse_variable(line):
                self.parse_line(line)
        if textblock_var is not None:
            raise InvalidOpenSesameScript(
                'Missing __end__ block for multiline variable {textblock_var}')

    def to_string(self, item_type=None):
        """
        Encodes the item into an OpenSesame definition string.

        Keyword arguments:
        item_type	--	The type of the item or None for autodetect.
                                        (default=None)

        Returns:
        The unicode definition string
        """
        if item_type is None:
            item_type = self.item_type
        s = u'define %s %s\n' % (item_type, self.name)
        for comment in self.comments:
            s += u'\t# %s\n' % comment.strip()
        for var in self.var:
            s += u'\t' + self.variable_to_string(var)
        return s

    def resolution(self):
        r"""Returns the display resolution and checks whether the resolution is
        valid.

        __Important note:__

        The meaning of 'resolution' depends on the
        back-end. For example,
        the legacy back-end changes the actual
        resolution of the display,
        whereas the other back-ends do not alter the
        actual display
        resolution, but create a 'virtual display' with the
        requested
        resolution that is presented in the center of the display.

        Returns
        -------
        tuple
            A (width, height) tuple
        """
        w = self.var.get(u'width')
        h = self.var.get(u'height')
        if type(w) != int or type(h) != int:
            raise InvalidValue(f'({w}, {h}) is not a valid resolution')
        return w, h

    def set_item_onset(self, time=None):
        r"""Set a timestamp for the onset time of the item's execution.

        Parameters
        ----------
        time, optional
            A timestamp or None to use the current time.

        Returns
        -------
        """
        if time is None:
            time = self.clock.time()
        self.experiment.var.set(u'time_%s' % self.name, time)
        return time

    def var_info(self):
        """Give a list of dictionaries with variable descriptions.

        Returns
        -------
        list
            A list of (variable, description) tuples
        """
        return [(u"time_%s" % self.name, u"[Timestamp of last item call]"),
                (u"count_%s" % self.name, u"[Number of item calls]")]
        
    def get(self, *args, **kwargs):
        raise IncompatibilityError(
            'Item.get() has been removed in OpenSesame 4. Please see '
            'the section in the documentation on using variables.')

    def set(self, *args, **kwargs):
        raise IncompatibilityError(
            'Item.set() has been removed in OpenSesame 4. Please see '
            'the section in the documentation on using variables.')


# alias for backwards compatibility
item = Item



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import warnings
from libopensesame.item import Item
from libopensesame.exceptions import PythonError, PythonSyntaxError, \
    AbortCoroutines


class InlineScript(Item):
    """Allows users to use Python code in their experiments."""
    
    description = u'Executes Python code'

    def reset(self):
        """See item."""
        self.var._prepare = u''
        self.var._run = u''
        self._var_info = None

    @property
    def workspace(self):

        return self.experiment.python_workspace

    def prepare(self):
        r"""Executes the prepare script. The code that you enter in the
        'prepare' tab of an inline_script item in the GUI is used as a body for
        this function.
        """
        super().prepare()
        # 'self' must always be registered, otherwise we get confusions between
        # the various inline_script items.
        self.workspace[u'self'] = self
        # Compile prepare script
        try:
            self.cprepare = self.workspace._compile(
                self.var.get(u'_prepare', _eval=False))
        except SyntaxError as e:
            raise PythonSyntaxError(
                'Syntax error in inline script (prepare phase)',
                line_nr=e.lineno)
        # Compile run script
        try:
            self.crun = self.workspace._compile(
                self.var.get(u'_run', _eval=False))
        except SyntaxError as e:
            raise PythonSyntaxError(
                'Syntax error in inline script (run phase)',
                line_nr=e.lineno)
        # Run prepare script
        try:
            self.workspace._exec(self.cprepare)
        except Exception as e:
            raise PythonError(
                'Error while executing inline script (prepare phase)')

    def run(self):
        r"""Executes the run script. The code that you enter in the 'run' tab
        of an inline_script item in the GUI is used as a body for this
        function.
        """
        self.set_item_onset()
        # 'self' must always be registered, otherwise we get confusions between
        # the various inline_script items.
        self.workspace[u'self'] = self
        try:
            self.workspace._exec(self.crun)
        except Exception as e:
            raise PythonError(
                'Error while executing inline script (run phase)')

    def coroutine(self, coroutines):
        """See coroutines plug-in"""
        yield
        self.set_item_onset()
        while True:
            self.workspace[u'self'] = self
            try:
                self.workspace._exec(self.crun)
            except AbortCoroutines as e:
                # If the inline_script is part of a coroutines, this signals
                # that the coroutines should be aborted, so we don't wrap it
                # into a PythonError.
                raise
            except Exception as e:
                raise PythonError(
                    'Error while executing inline script (coroutines)')
            yield


# Alias for backwards compatibility
inline_script = InlineScript



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets, QtCore
import webbrowser
import yaml
from libopensesame import metadata
from libqtopensesame.extensions import BaseExtension
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc.translate import translation_context
from urllib.request import urlopen
_ = translation_context(u'help', category=u'extension')


class ActionPage(QtWidgets.QAction, BaseSubcomponent):

    r"""A menu entry for a single help page."""
    def __init__(self, main_window, title, link, menu):
        r"""Constructor.

        Parameters
        ----------
        main_window
            The main-window object.
        title
            The menu title.
        link
            The URL to open.
        menu
            The menu for the action.
        """
        QtWidgets.QAction.__init__(
            self, main_window.theme.qicon(u'applications-internet'),
            title,
            menu
        )
        self.setup(main_window)
        self.title = title
        self.link = link
        self.triggered.connect(self.open_page)

    def open_page(self):
        r"""Opens the page URL."""
        webbrowser.open(self.link)


class GetSitemapThread(QtCore.QThread):

    r"""A thread that asynchronously retrieves the documentation sitemap."""
    def __init__(self, help, sitemap_url, local_sitemap):

        QtCore.QThread.__init__(self)
        self._help = help
        self._sitemap_url = sitemap_url
        self._local_sitemap = local_sitemap
        self.sitemap = None

    def run(self):

        try:
            fd = urlopen(self._sitemap_url)
            self.sitemap = fd.read()
        except Exception:
            if self._local_sitemap is None:
                return
            with safe_open(self._help.ext_resource(self._local_sitemap)) as fd:
                self.sitemap = fd.read()


class Help(BaseExtension):

    def event_startup(self):
        r"""Build the menu on startup."""
        self._urls = []
        self.menu = self.menubar.addMenu(_(u'Help'))
        self._wait_action = QtWidgets.QAction(
            self.theme.qicon(u'process-working'),
            _(u'Please wait …'),
            self.menu
        )
        self._wait_action.setEnabled(False)
        self.menu.addAction(self._wait_action)
        # It can take some time to download the sitemap from cogsci. Therefore,
        # we use threading.
        self._get_sitemap_thread = GetSitemapThread(
            self,
            sitemap_url=cfg.online_help_sitemap.replace(
                u'[version]',
                metadata.major_version
            ),
            local_sitemap=None
        )
        self._get_sitemap_thread.start()
        self._get_sitemap_thread.finished.connect(self.populate_help_menu)

    def populate_help_menu(self):
        r"""Is collected when the sitemap is done loading, and populates the
        menu.
        """
        if self._get_sitemap_thread.sitemap is None:
            return
        try:
            _dict = safe_yaml_load(self._get_sitemap_thread.sitemap)
        except yaml.scanner.ScannerError:
            # If the sitemap was loaded but was not a valid yaml string. This
            # can happen for example when a captive portal returns something
            # invalid
            return
        if not isinstance(_dict, dict):
            return
        self.menu.clear()
        menu = self.build_menu(
            self.menu,
            cfg.online_help_base_url,
            _(u'Online help'),
            _dict,
            sub_menu=cfg.online_help_psychopy
        )
        if menu is not None:
            self.action_online_help = self.menu.addMenu(menu)
        if cfg.online_help_psychopy:
            menu = self.psychopy_help_menu()
            if menu is not None:
                self.action_psychopy_help = self.menu.addMenu(menu)
        self.menu.addSeparator()

    def build_menu(self, parent_menu, base_url, title, _dict, sub_menu=True):
        r"""A helper function to build the online-help menu.

        Parameters
        ----------
        parent_menu
            A parent QMenu for a submenu.
        base_url
            A prefix for relative urls.
        title
            A menu title.
        _dict
            A dict extracted from the sitemap with page names and contents.

        Returns
        -------
        A QMenu.
        """
        if sub_menu:
            menu = parent_menu.addMenu(
                self.theme.qicon(u'applications-internet'),
                title
            )
        else:
            menu = parent_menu
        for name, link in _dict.items():
            if not isinstance(link, str):
                self.build_menu(menu, base_url, name, link)
                return
            if not self._is_url(link):
                link = base_url+link
            self._urls.append(link)
            menu.addAction(
                ActionPage(
                    self.main_window,
                    safe_decode(name, enc=u'utf-8'),
                    link,
                    menu
                )
            )
        return menu

    def psychopy_help_menu(self):
        r"""Builds a PsychoPy help menu.

        Returns
        -------
        A QMenu.
        """
        with safe_open(self.ext_resource(u'psychopy_sitemap.yaml')) as fd:
            sitemap = fd.read()
        _dict = safe_yaml_load(sitemap)
        menu = self.build_menu(self.menu, None, _(u'PsychoPy API'), _dict)
        return menu

    def _is_url(self, link):
        r"""Verifies whether a string corresponds to an http(s) url."""
        return link.startswith(u'http://') or link.startswith(u'https://')



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._canvas._element.element import Element


class Rect(Element):

    def __init__(self, canvas, x, y, w, h, **properties):

        properties = properties.copy()
        x, y, w, h = self._rect(x, y, w, h)
        properties.update({'x': x, 'y': y, 'w': w, 'h': h})
        Element.__init__(self, canvas, **properties)

    @property
    def rect(self):

        return self.x, self.y, self.w, self.h



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.widgets._form import Form
from openexp.canvas_elements import Rect


class Widget:

    r"""The base class for all other widgets."""
    def __init__(self, form):
        r"""Constructor.

        Parameters
        ----------
        form : form
            The parent form.
        """
        self.type = u'widget'
        self.form = form
        self.frame = False
        self.rect = None
        self._focus = False
        self.var = None
        # Check if the form parameter is valid
        if not isinstance(form, Form):
            raise TypeError(f'The first parameter passed to the constructor '
                            f'of a form widget should be a form, not {form}')

    @property
    def focus(self):

        return self._focus

    @focus.setter
    def focus(self, focus):

        self._focus = focus
        self._update()

    @property
    def box_size(self):
        return self.form.theme_engine.box_size()

    @property
    def theme_engine(self):
        return self.form.theme_engine

    @property
    def canvas(self):
        return self.form.canvas

    def _inside(self, pos, rect):

        x1, y1 = pos
        x2, y2, w, h = rect
        return not (x1 < x2 or y1 < y2 or x1 > x2+w or y1 > y2+h)

    def _init_canvas_elements(self):
        r"""Initializes all canvas elements."""
        self._frame_elements = {}

    def _update_frame(self, rect=None, style=u'normal'):
        r"""Draws a simple frame around the widget.

        Parameters
        ----------
        rect : tuple, NoneType, optional
            A (left, top, width, height) tuple for the frame geometry or `None`
            to use the widget geometry.
        style : str, unicode, optional
            A visual style. Should be 'normal', 'active', or 'light'.
        """
        if not self.frame:
            return
        if style not in self._frame_elements:
            element = self.theme_engine.frame(*self.rect, style=style)
            self._frame_elements[style] = element
            self.canvas.add_element(element)
            self.canvas.lower_to_bottom(element)
        for element_style, element in self._frame_elements.items():
            element.visible = element_style == style

    def _update(self):
        r"""Draws the widget."""
        self._update_frame(self.rect)

    def set_rect(self, rect):
        r"""Sets the widget geometry.

        Parameters
        ----------
        rect : tuple
            A (left, top, width, height) tuple.
        """
        self.rect = rect
        self._init_canvas_elements()
        self._update()

    def set_var(self, val, var=None):
        r"""Sets an experimental variable.

        Parameters
        ----------
        val    A value.
        var : str, unicode, NoneType, optional
            A variable name, or None to use widget default.
        """
        if var is None:
            var = self.var
        if var is None:
            return
        if isinstance(val, bool):
            val = 'yes' if val else 'no'
        self.form.experiment.var.set(
            self.form.experiment.syntax.eval_text(var), val)

    def on_key_press(self, key):
        r"""Is called whenever the widget is focused and the users enters a
        key.

        Parameters
        ----------
        key : str
            A key
        """
        pass

    def on_mouse_click(self, pos):
        r"""Is called whenever the user clicks on the widget.

        Parameters
        ----------
        pos : tuple
            An (x, y) coordinates tuple.
        """
        pass

    def coroutine(self):
        r"""Implements the interaction. This can be overridden to implement
        more complicated keyboard/ mouse interactions.
        """
        retval = None
        while True:
            d = yield retval
            if d[u'type'] == u'click':
                retval = self.on_mouse_click(d['pos'])
            elif d[u'type'] == u'key':
                retval = self.on_key_press(d[u'key'])
            elif d[u'type'] == u'stop':
                break


widget = Widget



"""A grid-based response item, convenient for touch screens"""

category = "Response collection"
icon = "os-touch_response"
controls = [
    {
        "type": "line_edit",
        "var": "correct_response",
        "label": "Correct response",
        "tooltip": "Set the correct response"
    },
    {
        "type": "line_edit",
        "var": "timeout",
        "label": "Timeout",
        "tooltip": "Expecting a value in milliseconds or 'infinite'"
    },
    {
        "type": "spinbox",
        "var": "_ncol",
        "label": "Number of columns",
        "min_val": 1,
        "max_val": 100,
        "tooltip": "Specifies the number of columns"
    },
    {
        "type": "spinbox",
        "var": "_nrow",
        "label": "Number of rows",
        "min_val": 1,
        "max_val": 100,
        "tooltip": "Specifies the number of rows"
    },
    {
        "type": "checkbox",
        "var": "show_cursor",
        "label": "Show cursor",
        "tooltip": "Show a mouse cursor (if supported on device)"
    }
]



"""Preloads items in the background for smoother use"""

icon = "applications-accessories"
label = "Preload items"
settings = {
    "preload_items_interval": 1000,
    "preload_items_idle_threshold": 0.01

}



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame import misc
from libopensesame.widgets._widget import Widget
from PIL import Image
from libopensesame.exceptions import ImageDoesNotExist, UnsupportedImageFormat
from openexp.canvas_elements import Image as ImageElement
import os
import sys


class ImageWidget(Widget):

    r"""The `Image` widget is used to display a non-interactive image.
    __Example (OpenSesame script):__

    ~~~
    widget 0 0 1 1 image path='5.png'
    ~~~
    __Example (Python):__

    ~~~ .python
    form = Form()
    # The full path to the
    image needs to be provided.
    # self.experiment.pool can be used to retrieve
    the full path
    # to an image in the file pool.
    image =
    ImageWidget(path=pool['5.png'])
    form.set_widget(image, (0,0))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(self, form, path=None, adjust=True, frame=False):
        r"""Constructor to create a new `ImageWidget` object. You do not
        generally call this constructor directly, but use the
        `ImageWidget()`
        factory function, which is described here:
        [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        path : str, unicode, NoneType, optional
            The full path to the image. To show an image from the file pool,
            you need to first use `experiment.get_file` to determine the full
            path to the image.
        adjust : bool, optional
            Indicates whether the image should be scaled according to the size
            of the widget.
        frame : bool, optional
            Indicates whether a frame should be drawn around the widget.
        """
        if isinstance(adjust, str):
            adjust = adjust == u'yes'
        if isinstance(frame, str):
            frame = frame == u'yes'
        Widget.__init__(self, form)
        self.adjust = adjust
        self.frame = frame
        self.path = path
        self.type = u'image'

    def _init_canvas_elements(self):
        r"""Initializes all canvas elements."""
        _path = safe_str(self.path, enc=sys.getfilesystemencoding())
        if not os.path.exists(_path):
            raise ImageDoesNotExist(_path)
        x, y, w, h = self.rect
        x += w/2
        y += h/2
        self.canvas.add_element(
            ImageElement(_path, x=x, y=y, scale=self.scale, center=True)
            .construct(self.canvas)
        )
        Widget._init_canvas_elements(self)
        if self.frame:
            self._update_frame(self.rect)

    def set_rect(self, rect):
        r"""Sets the widget geometry.

        Parameters
        ----------
        rect : tuple
            A (left, top, width, height) tuple.
        """
        self.rect = rect
        _path = safe_str(self.path, enc=sys.getfilesystemencoding())
        if not os.path.isfile(_path):
            raise ImageDoesNotExist(_path)
        if self.adjust:
            x, y, w, h = self.rect
            try:
                img = Image.open(_path)
                img_w, img_h = img.size
            except:
                try:
                    import pygame
                    img = pygame.image.load(_path)
                except:
                    raise UnsupportedImageFormat(_path)
                img_w, img_h = img.get_size()
            scale_x = 1.*w/img_w
            scale_y = 1.*h/img_h
            self.scale = min(scale_x, scale_y)
        else:
            self.scale = 1
        Widget.set_rect(self, rect)


image = ImageWidget



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue, MissingDependency
from openexp.sampler import Sampler
try:
    import numpy as np
    from scipy import signal
except ImportError:
    np = None
    signal = None


def Synth(experiment, osc="sine", freq=440, length=100, attack=0, decay=5):
    r"""A factory that synthesizes a sound and returns it as a `sampler
    object`. For a full description of keywords, see
    `python_workspace_api.synth`.

    For backwards compatibility, this function behaves as though it is a
    back-end.

    Parameters
    ----------
    experiment : experiment
        The experiment object.

    Returns
    -------
    sampler
        A SAMPLER object.
    """
    if np is None:
        raise MissingDependency(
            'The synth is not available, because numpy is missing.')
    if attack < 0 or attack > length:
        raise InvalidValue(
            'Attack must be a numeric value between 0 and the sound length')
    if decay < 0 or decay > length:
        raise InvalidValue(
            'Decay must be a numeric value between 0 and the sound length')
    # We need to multiply the rate by two to get a stereo signal
    rate = 2*experiment.var.get('sound_freq', 48000)
    signal = osc_gen(osc, key_to_freq(freq), length, rate)
    _envelope = envelope(length, attack, decay, rate)
    sound = to_int_16(signal * _envelope)
    return Sampler(experiment, sound)


def key_to_freq(key):
    r"""Converts a key (e.g., A1) to a frequency.

    Parameters
    ----------
    key : str, unicode, int, float
        A string like "A1", "eb2", etc, or a numeric frequency (in which case
        the frequency is simply returned as a float).

    Returns
    -------
    float
        A frequency in hertz.

    Examples
    --------
    >>> from openexp.synth import synth
    >>> my_synth = synth(exp)
    >>> print('An a2 is %d Hz' % my_synth.key_to_freq('a2'))
    """
    if type(key) in [int, float]:
        return key
    if not isinstance(key, str) or len(key) < 2:
        raise InvalidValue(
            "synth.key_to_freq(): '%s' is not a valid note, expecting something like 'A1'")
    n = key[:-1].lower()
    try:
        o = int(key[-1])
    except:
        raise InvalidValue(
            "synth.key_to_freq(): '%s' is not a valid note, expecting something like 'A1'")
    if n == "a":
        f = 440.0
    elif n == "a#" or n == "bb":
        f = 466.16
    elif n == "b":
        f = 493.92
    elif n == "c":
        f = 523.28
    elif n == "c#" or n == "db":
        f = 554.40
    elif n == "d":
        f = 587.36
    elif n == "d#" or n == "eb":
        f = 698.47
    elif n == "e":
        f = 659.48
    elif n == "f":
        f = 698.48
    elif n == "f#" or n == "gb":
        f = 740.00
    elif n == "g":
        f = 784.00
    elif n == "ab" or n == "g#":
        f == 830.64
    if o < 1:
        o = 0.5 ** (abs(o) + 1)
        freq = f * o
    else:
        freq = f * o
    return freq


def osc_gen(_type, freq, length, rate):

    length *= .001
    t = np.linspace(0, length, int(length*rate))
    if _type == 'square':
        return signal.square(2 * np.pi * freq * t)
    if _type == 'saw':
        return signal.sawtooth(2 * np.pi * freq * t)
    if _type == 'sine':
        return np.sin(2 * np.pi * freq * t)
    if _type == 'white_noise':
        return np.random.random(int(length * rate)) * 2 - 1
    raise InvalidValue('Invalid oscillator: %s' % _type)


def envelope(length, attack, decay, rate):
    r"""An envelope generator that determines the volume profile of the sound."""
    length *= .001
    attack *= .001
    decay *= .001
    e = np.ones(int(length*rate))
    if attack > 0:
        attack = int(attack*rate)
        e[:attack] *= np.linspace(0, 1, attack)
    if decay > 0:
        decay = int(decay*rate)
        e[-decay:] *= np.linspace(1, 0, decay)
    return e


def to_int_16(a):
    r"""Converts the float array to an 16 bit int array, which is a more
    typical sound format.
    """
    a *= 32767
    return a.astype(np.int16)


# Non PEP-8 alias for backwards compatibility
synth = Synth



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtGui, QtCore, QtWidgets
from libqtopensesame.misc.config import cfg
from libqtopensesame.misc import drag_and_drop
from libqtopensesame.misc.shortcut import Shortcut
from libqtopensesame.misc.base_subcomponent import BaseSubcomponent
from libqtopensesame.misc.base_draggable import BaseDraggable
from libqtopensesame.misc.run_if_delegate import RunIfDelegate
from libqtopensesame.misc.item_name_delegate import ItemNameDelegate
from libqtopensesame.widgets.tree_append_button import TreeAppendButton
from libqtopensesame._input.popup_menu import PopupMenu
from libqtopensesame.items.qtstructure_item import QtStructureItem
from libopensesame.oslogging import oslogger
from libopensesame.sequence import Sequence
from libqtopensesame.misc.translate import translation_context
_ = translation_context('tree_overview', category='core')


class TreeOverview(BaseSubcomponent, BaseDraggable, QtWidgets.QTreeWidget):

    """A tree widget used in sequence items and the overview area."""
    structure_change = QtCore.Signal()
    text_change = QtCore.Signal()

    NAVIGATION_KEYS = [
        QtCore.Qt.Key_Up,
        QtCore.Qt.Key_Down,
        QtCore.Qt.Key_PageUp,
        QtCore.Qt.Key_PageDown,
        QtCore.Qt.Key_Home,
        QtCore.Qt.Key_End,
        QtCore.Qt.Key_Return
    ]

    def __init__(self, parent, overview_mode=True):
        """Constructor.

        Parameters
        ----------
        parent : qtopensesame
            The main window object.
        overview_mode : int, optional
            Indicates whether the tree should be overview-area style (True) or
            sequence style (False).
        """
        super().__init__(parent)
        self.locked = False
        self.overview_mode = overview_mode
        self.setAcceptDrops(True)
        self.setItemDelegateForColumn(0, ItemNameDelegate(self))
        if self.overview_mode:
            self.setHeaderHidden(True)
        else:
            self.setItemDelegateForColumn(1, RunIfDelegate(self))
            self.setHeaderHidden(False)
            self.setHeaderLabels([_('Item name'), _('Run if')])
        self.setAlternatingRowColors(True)
        self.itemChanged.connect(self.text_edited)
        self.pending_drag_data = None
        self.drag_timer = None
        self.inhibit_drag = False
        if not self.overview_mode:
            Shortcut(self, cfg.shortcut_edit_runif, self.start_edit_runif)
            self.append_button = TreeAppendButton(self)
        else:
            self.append_button = None
        Shortcut(self, cfg.shortcut_context_menu, self.show_context_menu)
        Shortcut(self, cfg.shortcut_rename, self.start_rename)
        Shortcut(self, cfg.shortcut_copy_clipboard_unlinked,
                 self.copy_item_unlinked)
        Shortcut(self, cfg.shortcut_copy_clipboard_linked,
                 self.copy_item_linked)
        Shortcut(self, cfg.shortcut_paste_clipboard, self.paste_item)
        Shortcut(self, cfg.shortcut_delete, self.delete_item)
        Shortcut(self, cfg.shortcut_permanently_delete,
                 self.permanently_delete_item)
        self.drop_indicator = None
        self.drop_indicator_pen = QtGui.QPen(QtGui.QBrush(
            QtGui.QColor('#73d216')), 2, QtCore.Qt.SolidLine)
        self.set_supported_drop_types(['item-snippet', 'item-existing',
                                       'url-local'])

    def items(self):

        l = []
        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            l += [item] + item.children()
        return l

    def default_fold_state(self):

        self.topLevelItem(0).expand()
        if self.topLevelItemCount() > 0:
            self.topLevelItem(1).collapse()

    def get_fold_state(self):

        fold_state = {}
        for item in self.items():
            fold_state[item.path()] = item.isExpanded()
        return fold_state

    def set_fold_state(self, fold_state):

        for item in self.items():
            a = item.path()
            if a in fold_state:
                item.setExpanded(fold_state[a])
            else:
                item.expand()

    def show_context_menu(self):

        item = self.currentItem()
        if item is None:
            return
        index = self.indexFromItem(item)
        rect = self.visualRect(index)
        item.show_context_menu(self.mapToGlobal(rect.topLeft()))

    def copy_item_unlinked(self):
        """Copies the currently selected treeitem to the clipboard (if
        supported by the treeitem) as an unlinked copy.
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.copy_unlinked()

    def copy_item_linked(self):
        """Copies the currently selected treeitem to the clipboard (if
        supported by the treeitem) as a linked copy.
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.copy_linked()

    def paste_item(self):
        """Pastes the clipboard onto the currently selected treeitem (if
        supported by the treeitem).
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.paste()

    def delete_item(self):
        """Deletes the currently selected treeitem (if supported by the
        treeitem).
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.delete()

    def permanently_delete_item(self):
        """Permanently deletes the currently selected treeitem (if supported
        by the treeitem).
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.permanently_delete()

    def create_linked_copy(self):
        """Creates a linked copy of the currently selected item (if supported
        by the treeitem).
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.create_linked_copy()

    def create_unlinked_copy(self):
        """Creates an unlinked copy of the currently selected item (if
        supported by the treeitem).
        """
        target_treeitem = self.currentItem()
        if target_treeitem is not None:
            target_treeitem.create_unlinked_copy()

    def select_item(self, name, open_tab=True):
        """Selects all items that match the specified name.

        Parameters
        ----------
        name
            The name to select.
        """
        self.clearSelection()
        l = self.findItems(name,
                           QtCore.Qt.MatchFixedString | QtCore.Qt.MatchRecursive, 0)
        if len(l) == 0:
            return
        for _l in l:
            _l.setSelected(True)
        if open_tab:
            self.experiment.items[name].open_tab(select_in_tree=False)

    def text_edited(self, treeitem, col):
        """Processes edits to the item names and run-if statements.

        Parameters
        ----------
        treeitem : QTreeWidgetItem
            A tree item
        col : int
            The column that was edited.
        """
        if col == 0:
            if hasattr(treeitem, 'name'):
                from_name = treeitem.name
                to_name = self.experiment.items.rename(
                    from_name, treeitem.text(0))
                if to_name is None:
                    self.itemChanged.disconnect()
                    treeitem.setText(0, from_name)
                    self.itemChanged.connect(self.text_edited)
                    to_name = from_name
                if from_name != to_name:
                    self.itemChanged.disconnect()
                    treeitem.setText(0, to_name)
                    self.itemChanged.connect(self.text_edited)
                treeitem.name = to_name
                self.text_change.emit()
        elif col == 1:
            # Don't allow editing the run-if statement of the top-level
            # sequence if we are in non-overview mode.
            if treeitem.parent() is None:
                self.itemChanged.disconnect()
                treeitem.setText(1, '')
                self.itemChanged.connect(self.text_edited)
                return
            if hasattr(treeitem, 'ancestry'):
                parent_item_name, ancestry = treeitem.ancestry()
                parent_item_name, index = self.parent_from_ancestry(ancestry)
                parent_item = self.experiment.items[parent_item_name]
                if parent_item.item_type == 'sequence':
                    self.itemChanged.disconnect()
                    cond = treeitem.set_extra_info(treeitem.text(1).strip())
                    parent_item.set_run_if(index, cond)
                    self.itemChanged.connect(self.text_edited)
                self.text_change.emit()

    def mousePressEvent(self, e):
        """Initiates a drag event after a mouse press, these are used to
        implement item moving, shallow copying, and deep copying.

        Parameters
        ----------
        e : QMouseEvent
            A mouse event.
        """
        super().mousePressEvent(e)
        if e.buttons() != QtCore.Qt.LeftButton:
            return
        # Get item and open tab.
        target_treeitem = self.itemAt(e.pos())
        if target_treeitem is None:
            return
        # When opening a tab takes a while, a drag may be inadvertently
        # started. For that reason, items can set the inhibit_drag flag to
        # True so that a drag is inhibited, but only once.
        self.inhibit_drag = False
        if self.overview_mode:
            target_treeitem.open_tab()
        # Only start drags for draggable tree items.
        if self.inhibit_drag or not self.draggable(target_treeitem):
            e.ignore()
            return
        # Get the target item
        target_item_name, target_item_ancestry = target_treeitem.ancestry()
        # Automated bug reports suggest that target_item_name can sometimes be
        # an empty string. How this is possible is unclear, but this should
        # avoid these mysterious crashes.
        if not target_item_name:
            e.ignore()
            return
        target_item = self.experiment.items[target_item_name]
        data = {
            'type': 'item-existing',
            'move': True,
            'item-name': target_item.name,
            'item-type': target_item.item_type,
            'structure-item': isinstance(target_item, QtStructureItem),
            'ancestry': target_item_ancestry,
            'script': target_item.to_string(),
            'application-id': self.main_window._id(),
            'QTreeWidgetItem': str(target_treeitem),
        }
        # Drags are not started right away, but only after the mouse has been
        # pressed for a minimum interval. To accomplish this, we set a timer,
        # and cancel the timer when the mouse cursor is released.
        if self.drag_timer is not None:
            self.drag_timer.stop()
        self.pending_drag_data = data
        self.drag_timer = QtCore.QTimer()
        self.drag_timer.setSingleShot(True)
        self.drag_timer.setInterval(cfg.start_drag_delay)
        self.drag_timer.timeout.connect(self.start_drag)
        self.drag_timer.start()

    def start_drag(self):
        """Starts a pending drag operation (if any)."""
        if self.pending_drag_data is None:
            return
        drag_and_drop.send(self, self.pending_drag_data)
        self.end_drag()
        self.pending_drag_data = None

    def end_drag(self):
        """Ends a drag operation."""
        self.drop_indicator = None
        self.viewport().update()

    def mouseReleaseEvent(self, e):
        """Cancels pending drag operations when the mouse is released to
        quickly. This avoids accidental dragging.

        Parameters
        ----------
        e : QMouseEvent
            A mouse event.
        """
        super().mouseReleaseEvent(e)
        self.pending_drag_data = None

    def parent_from_ancestry(self, ancestry):
        """Gets the parent item, and the index of the item in the parent,
        based on an item's ancestry, as returned by ancestry_from_treeitem.

        Parameters
        ----------
        ancestry : unicode
            An ancestry string.

        Returns
        -------
        tuple
            A (parent item name, index) tuple.
        """
        l = ancestry.split('.')
        if len(l) == 1:
            return None, None
        parent_item_name = l[1].split(':')[0]
        index = int(l[0].split(':')[1])
        return parent_item_name, index

    def droppable(self, treeitem, data):
        """Determines if a tree item is able to accept drops.

        Parameters
        ----------
        treeitem : QTreeWidgetItem
            A tree item.
        data : dict
            Drag-and-drop data.

        Returns
        -------
        bool
            A bool indicating if the tree item is droppable.
        """
        return treeitem is not None and treeitem.droppable(data)

    def draggable(self, treeitem):
        """Determines if a tree item can be dragged.

        Parameters
        ----------
        treeitem : QTreeWidgetItem
            A tree item.

        Returns
        -------
        bool
            A bool indicating if the tree item is draggable.
        """
        return treeitem is not None and treeitem.draggable()

    def drop_event_item_existing(self, data, e=None, target_treeitem=None):
        """Handles drop events for item moves.

        Parameters
        ----------
        data: dict
            A drop-data dictionary
        e: QDropEvent, optional
        target_tree_item: TreeItemItem, optional
        """
        if not drag_and_drop.matches(data, ['item-existing']):
            if e is not None:
                e.ignore()
            return
        # Only accept existing-item drops from this application
        if data['application-id'] != self.main_window._id():
            oslogger.debug('Drop ignored: from different instance')
            if e is not None:
                e.ignore()
            return
        # Determine the tree item that is dropped onto
        if target_treeitem is None:
            try:
                pos = e.pos()
            except AttributeError:
                # This seems to be a backwards-incompatibility bug in PyQt6,
                # which has not implemented the pos() function
                pos = e.position().toPoint()
            target_treeitem = self.itemAt(pos)
        if not self.droppable(target_treeitem, data):
            oslogger.debug('Drop ignored: target not droppable')
            if e is not None:
                e.ignore()
            return
        # Don't drop on the same item that was the source (if any)
        if 'QTreeWidgetItem' in data and \
                data['QTreeWidgetItem'] == str(target_treeitem):
            oslogger.debug('Drop ignored: dropping on self')
            if e is not None:
                e.ignore()
            return
        item_name = data['item-name']
        # Check for recursion when dropping sequences and loops. The target
        # item may not have the dropped item in its ancestry. However, the
        # target item may occur multiple times in the experiment, so we need
        # to check that this constraint holds for all linked copies of the
        # target item.
        if data.get('structure-item', False):
            for linked_target_treeitem in self.findItems(
                    target_treeitem.name,
                    QtCore.Qt.MatchFixedString | QtCore.Qt.MatchRecursive):
                target_item_name, target_item_ancestry = \
                    linked_target_treeitem.ancestry()
                if target_item_ancestry.startswith('%s:' % item_name) \
                        or '.%s:' % item_name in target_item_ancestry:
                    oslogger.debug('Drop ignored: recursion prevented')
                    if e is not None:
                        e.ignore()
                    return
            # If the dropped item is in the unused items bin, then we need to
            # check whether the target item is a child.
            if item_name in self.unused_items():
                if target_treeitem.name in \
                        self.experiment.items[item_name].children():
                    if e is not None:
                        e.ignore()
                    return
        # Don't drop on undroppable parents
        parent_item_name, index = self.parent_from_ancestry(data['ancestry'])
        if parent_item_name is None:
            oslogger.debug('Drop ignored: no parent')
            if e is not None:
                e.ignore()
            return
        # * The actual drop action starts below *
        #
        # The logic below is a bit complicated, but works as follows:
        # - If we're in a move action, remove the dragged item from its parent,
        #   and set need_restore so that we know this happened.
        # - Try to drop the dragged item onto the target item
        # - If the drop action was unsuccessful, and if need_restore is set,
        #   re-add the dragged item to its former parent.
        event_action = 'move_item' if data['move'] else 'linked_copy'
        self.extension_manager.fire(f'prepare_{event_action}', name=item_name)
        need_restore = False
        if data['move']:
            if parent_item_name not in self.experiment.items:
                oslogger.debug(
                    f'Don\'t know how to remove item from {parent_item_name}')
            else:
                self.locked = True
                need_restore = True
                self.experiment.items[parent_item_name].remove_child_item(
                    item_name, index)
                self.locked = False
        if not self.drop_event_item_new(
                data, e, target_treeitem=target_treeitem) and need_restore:
            self.experiment.items[parent_item_name].insert_child_item(
                item_name, index)
            self.experiment.build_item_tree()
        self.extension_manager.fire(event_action, name=item_name)

    def drop_get_item_snippet(self, data):
        """Gets the item and list of newly created items for item-snippet
        drops.

        Parameters
        ----------
        data : dict
            The drop data.

        Returns
        -------
        tuple
            An (name, new_items) tuple.
        """
        for item_dict in data['items']:
            if not self.experiment.items.valid_type(item_dict['item-type']):
                raise TypeError(
                    _('Unknown item type: %s') % item_dict['item-type'])
        rename = []
        new_items = []
        main_item = None
        # Loop through all to-be-created items and instantiate them based on
        # the type, name, and script. The script is empty for newly created
        # items, but not for unlinked copies. The name functions as a
        # suggestion, which will be ignored if the suggested name is already
        # taken. In that case, we remember all the renamings and inform all
        # newly created items of this at the very end.
        for item_dict in data['items']:
            item = self.experiment.items.new(item_dict['item-type'],
                                             item_dict['item-name'],
                                             item_dict['script'],
                                             catch_exceptions=False)
            if item_dict['item-name'] == data['main-item-name']:
                main_item = item
            # If the item didn't get the suggested name
            if item.name != item_dict['item-name']:
                rename.append((item_dict['item-name'], item.name))
            new_items.append(item)
            self.extension_manager.fire('new_item', name=item.name,
                                        _type=item.item_type)
        # Inform all newly created items of any renames that occurred
        for old_name, new_name in rename:
            for item in new_items:
                item.rename(old_name, new_name)
        return main_item, [item.name for item in new_items]

    def drop_get_item_existing(self, data):
        """Gets the item and list of newly created items for existing-item
        drops.

        Parameters
        ----------
        data : dict
            The drop data.

        Returns
        -------
        tuple
            An (name, new_items) tuple, where new_items is always empty for
            existing-item drops.
        """
        if data['item-name'] not in self.experiment.items:
            raise NameError('Unknown item: %s' % data['item-name'])
        return self.experiment.items[data['item-name']], []

    def drop_event_item_new(self, data, e=None, target_treeitem=None):
        """Handles drop events for item creation.

        Parameters
        ----------
        data : dict
            A drop-data dictionary.
        e : QDropEvent or None, optional
            A drop event or None if a target treeitem is provided.
        target_treeite : TreeBaseItem or None, optional
            A target tree item or None in a drop event is specified.

        Returns
        -------
        bool
            True if the drop was successful, False otherwise.
        """
        self.main_window.set_busy(True)
        if not drag_and_drop.matches(data,
                                     ['item-snippet', 'item-existing']):
            if e is not None:
                e.ignore()
            self.main_window.set_busy(False)
            return False
        # Check that we're not trying to make a linked copy of a non-existent
        # item. This could happen if the user first performs a linked-copy
        # action, then deletes the copied item, and finally tries to paste the
        # linked copy.
        if data['type'] == 'item-existing' and \
                data['item-name'] not in self.experiment.items:
            self.notify(
                _('Cannot create linked copy of "%s". Has '
                  'the item been permanently deleted?') % data['item-name'])
            if e is not None:
                e.ignore()
            self.main_window.set_busy(False)
            return False
        # Determine the tree item on which the drop is performed. While doing
        # so we take into account what seems to be a backwards-incompatibility
        # bug in PyQt6, which has not implemented the pos() function
        if target_treeitem is None:
            try:
                pos = e.pos()
            except AttributeError:
                pos = e.position().toPoint()
            target_treeitem = self.itemAt(pos)
        # Ignore drops on non-droppable tree items.
        if not self.droppable(target_treeitem, data):
            if e is not None:
                e.ignore()
            self.main_window.set_busy(False)
            return False
        # Accept drops on the unused items bin and unused items (i.e. items
        # in the bin)
        if target_treeitem.name == '__unused__' or \
                (target_treeitem.parent() is not None and
                 target_treeitem.parent().name == '__unused__'):
            e.accept()
            self.structure_change.emit()
            self.main_window.set_busy(False)
            return True
        # Get the target item, and ignore the drop action if the target item
        # doesn't exist. (Note that the target item refers to an Item object
        # whereas the target treeitem refers to a TreeItemItem object that
        # represents an Item object in the overview area.)
        target_item_name = target_treeitem.text(0)
        if target_item_name not in self.experiment.items:
            oslogger.debug('Don\'t know how to drop on %s' % target_item_name)
            if e is not None:
                e.ignore()
            self.structure_change.emit()
            self.main_window.set_busy(False)
            return False
        target_item = self.experiment.items[target_item_name]
        # Perform the drop action, which results in an (item name, new_items)
        # tuple, where the item name is the item to be insert into the target
        # item, and the new_items is a list of newly created Items objects,
        # which may be empty (in the case of linked copies)
        if data['type'] == 'item-existing':
            # Perform a linked copy.
            item, new_items = self.drop_get_item_existing(data)
        else:
            # Perform an unlinked copy.
            # Creating an item may fail, and we therefore need to clean up when
            # this happens. But we don't catch the Exception itself, because it
            # will be handled with higher up, for example by the bug_report
            # extension.
            try:
                item = None
                item, new_items = self.drop_get_item_snippet(data)
            finally:
                if item is None:
                    if e is not None:
                        e.ignore()
                    self.structure_change.emit()
                    self.main_window.set_busy(False)
                    self.end_drag()
        inserted = False
        # If the item has no parent or if it is the experiment starting point,
        # we insert into it directly.
        if target_treeitem.parent() is None or \
                target_item.name == self.experiment.var.start:
            target_item.insert_child_item(item.name)
            inserted = True
        else:
            if isinstance(target_item, QtStructureItem):
                self.main_window.set_busy(False)
                resp = PopupMenu(self, [
                    (0, _('Insert into %s') % target_item.name, 'go-next'),
                    (1, _('Insert after %s') %
                     target_item.name, 'go-down')
                ]).show()
                # If the popup was cancelled
                if resp is None:
                    if e is not None:
                        e.accept()
                    self.main_window.set_busy(False)
                    for item in new_items:
                        del self.experiment.items[item]
                    return False
                # If the user chose to insert into the target item
                if resp == 0:
                    target_item.insert_child_item(item.name)
                    inserted = True
        # Otherwise, we find the parent of the target item, and insert the
        # new item at the correct position.
        if not inserted:
            while True:
                try:
                    parent_treeitem = target_treeitem.parent()
                except:
                    # A race condition can occur in which the tree_overview has
                    # been rebuild, thus destroying target_treeitem. If this
                    # happens, we re-take target_treeitem based on the mouse
                    # coordinates.
                    target_treeitem = self.itemAt(e.pos())
                    parent_treeitem = target_treeitem.parent()
                if parent_treeitem is None:
                    e.accept()
                    del self.experiment.items[item.name]
                    self.main_window.set_busy(False)
                    return False
                parent_item_name = str(parent_treeitem.text(0))
                parent_item = self.experiment.items[parent_item_name]
                if isinstance(parent_item, Sequence):
                    break
                target_treeitem = parent_treeitem
            index = parent_treeitem.indexOfChild(target_treeitem) + 1
            # Below we correct the index when an item is moved down in the same
            # sequence. In this case, the item is first removed from the
            # sequence (in drop_event_item_existing) and because the treewidget
            # is not refreshed, the index as returned by the parent_treeitem
            # is one too high. (This is clearly a hack, but it seems to work.)
            children = parent_item.direct_children()
            if len(children) > 1:
                try:
                    if (
                            children[index - 1] != target_item.name and
                            children[index - 2] == target_item.name
                    ):
                        index -= 1
                except IndexError:
                    pass
            parent_item.insert_child_item(item.name, index=index)
        if e is not None:
            e.accept()
        self.structure_change.emit()
        if self.overview_mode:
            item.open_tab()
        self.main_window.set_busy(False)
        return True

    def dropEvent(self, e):
        """Accept drops.

        Parameters
        ----------
        e : QDragLeaveEvent
            A drop event.
        """
        # The focusOutEvent arrives too late, so here we explicitly give the
        # currently opened item the opportunity to get ready
        item_name = self.tabwidget.current_item()
        if item_name is not None:
            self.item_store[item_name].get_ready()
        data = drag_and_drop.receive(e)
        if data['type'] == 'item-snippet':
            self.drop_event_item_new(data, e)
        elif data['type'] == 'item-existing':
            self.drop_event_item_existing(data, e)
        elif data['type'] == 'url-local':
            self.main_window.open_file(path=data['url'])
            e.accept()
        else:
            e.ignore()
        self.end_drag()

    def dragLeaveEvent(self, e):
        """Cancels the drop indicator when a drag leaves.

        Parameters
        ----------
        e : QDragLeaveEvent
            A drag-move event.
        """
        self.end_drag()

    def dragMoveEvent(self, e):
        """Handles drag-move events to see if the item tree can handle
        incoming drops.

        Parameters
        ----------
        e : QDragMoveEvent
            A drag-move event.
        """
        data = drag_and_drop.receive(e)
        self.drop_indicator = None
        if drag_and_drop.matches(data, ['url-local']):
            e.accept()
            self.end_drag()
            return
        if not drag_and_drop.matches(data, ['item-snippet', 'item-existing']):
            e.accept()
            self.end_drag()
            return
        try:
            pos = e.pos()
        except AttributeError:
            # This seems to be a backwards-incompatibility bug in PyQt6,
            # which has not implemented the pos() function
            pos = e.position().toPoint()
        target = self.itemAt(pos)
        if not self.droppable(target, data):
            self.end_drag()
            e.ignore()
            return
        e.accept()
        # Update the drop indicator
        index = self.indexFromItem(target)
        rect = self.visualRect(index)
        if target.name == '__unused__' or (
                target.item.name in self.experiment.items.used() and
                isinstance(target.item, QtStructureItem) and
                target.item.name != self.experiment.var.start and
                target.parent() is not None):
            self.drop_indicator = rect
        else:
            self.drop_indicator = QtCore.QRect(rect.left(), rect.bottom(),
                                               rect.width(), 0)
        self.viewport().update()

    def paintEvent(self, e):
        """A custom pain event for the drop_indicator.

        Parameters
        ----------
        e : QPaintEvent
        """
        painter = QtGui.QPainter(self.viewport())
        self.drawTree(painter, e.region())
        if self.drop_indicator is not None:
            painter.setPen(self.drop_indicator_pen)
            if self.drop_indicator.height() == 0:
                painter.drawLine(self.drop_indicator.topLeft(),
                                 self.drop_indicator.topRight())
            else:
                painter.drawRect(self.drop_indicator)

    def contextMenuEvent(self, e):
        """Asks a tree item to show a context menu.

        Parameters
        ----------
        e : QContextMenuEvent
        """
        item = self.itemAt(e.pos())
        if item is not None:
            item.show_context_menu(e.globalPos())

    def keyPressEvent(self, e):
        """
        Capture key presses to auto-activate selected items

        Arguments:
        e -- a QKeyEvent
        """
        super().keyPressEvent(e)
        current_item = self.currentItem()
        if current_item is None:
            return
        # In overview mode, navigating through the items causes tabs to be
        # opened
        if self.overview_mode:
            if e.key() in self.NAVIGATION_KEYS:
                current_item.open_tab()
            return
        # In sequence view, the run-if statements can be changed using
        # Control+Plus (always) and Control+Minus (never)
        if QtCore.Qt.ControlModifier & e.modifiers():
            if e.key() == QtCore.Qt.Key_Minus:
                current_item.setText(1, 'never')
            elif e.key() in [QtCore.Qt.Key_Plus, QtCore.Qt.Key_Equal]:
                current_item.setText(1, 'always')

    def recursive_children(self, item):
        """
        Get a list of unused items from the itemtree

        Returns:
        A list of items (strings) that are children of the parent item
        """
        children = []
        for i in range(item.childCount()):
            child = item.child(i)
            children.append(str(child.text(0)))
            children += self.recursive_children(child)
        return children

    def unused_items(self):
        """
        Get a list of unused items

        Returns:
        A list of unused items (names as strings)
        """
        return self.main_window.ui.itemtree.recursive_children(
            self.main_window.ui.itemtree.topLevelItem(
                self.main_window.ui.itemtree.topLevelItemCount()-1))

    def rename(self, from_name, to_name):
        """Renames an item.

        Parameters
        ----------
        from_name : unicode
            The old name.
        to_name : unicode
            The new name.
        """
        self.itemChanged.disconnect()
        for i in range(self.topLevelItemCount()):
            self.topLevelItem(i).rename(from_name, to_name)
        self.itemChanged.connect(self.text_edited)

    def set_icon(self, name, icon):
        """Changes an item's icon.

        Parameters
        ----------
        name : unicode
            The item name.
        icon : unicode
            The icon name.
        """
        self.itemChanged.disconnect()
        for i in range(self.topLevelItemCount()):
            self.topLevelItem(i).set_icon(name, icon)
        self.itemChanged.connect(self.text_edited)

    def start_edit_runif(self):
        """Edits the run-if statement. This is not applicable in overview
        mode.
        """
        if self.overview_mode:
            return
        target_treeitem = self.currentItem()
        if target_treeitem is None:
            return
        if target_treeitem.parent() is None:
            # The top-level item is the sequence itself, which doesn't have a
            # run-if statement.
            return
        self.editItem(target_treeitem, 1)

    def start_rename(self):
        """Edits the item name."""
        target_treeitem = self.currentItem()
        if target_treeitem is not None \
                and target_treeitem.flags() & QtCore.Qt.ItemIsEditable:
            self.editItem(target_treeitem, 0)

    def setup(self, main_window):
        """This function needs to be overridden so that the append button is
        also set up.
        """
        super().setup(main_window)
        if self.append_button is not None:
            self.append_button.setup(main_window)

    def clear(self):
        """If the tree is cleared, we need to unset the target tree item in
        the append menu (if any).
        """
        super().clear()
        if self.append_button is None:
            return
        self.append_button.append_menu.target_treeitem = None

    def focusInEvent(self, e):
        """Select the general tab if no item is currently selected."""
        super().focusInEvent(e)
        if len(self.selectedItems()) == 0:
            self.setCurrentItem(self.topLevelItem(0))


# Alias for backwards compatibility
tree_overview = TreeOverview



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libopensesame.sequence import Sequence as SequenceRuntime
from libqtopensesame.widgets.tree_item_item import TreeItemItem
from libqtopensesame.widgets.tree_overview import TreeOverview
from libqtopensesame.items.qtitem import requires_init
from libqtopensesame.items.qtplugin import QtPlugin
from libqtopensesame.items.qtstructure_item import QtStructureItem
from libqtopensesame.misc.translate import translation_context
_ = translation_context(u'sequence', category=u'item')


class Sequence(QtStructureItem, QtPlugin, SequenceRuntime):
    """GUI controls for the sequence item.
    
    Parameters
    ----------
    name
        The item name.
    experiment
        The experiment object.
    string, optional
        A definition string.
    """
    
    description = _(u'Runs a number of items in sequence')
    help_url = u'manual/structure/sequence'
    lazy_init = True

    def __init__(self, name, experiment, string=None):
        SequenceRuntime.__init__(self, name, experiment, string)
        QtStructureItem.__init__(self)
        QtPlugin.__init__(self)
        self.last_removed_child = None, None

    def init_edit_widget(self):
        super().init_edit_widget(False)
        self.checkbox_flush_keyboard = QtWidgets.QCheckBox(
            _(u'Flush pending key presses at sequence start'))
        self.checkbox_flush_keyboard.setToolTip(
            _(u'Flush pending key presses at sequence start'))
        self.auto_add_widget(self.checkbox_flush_keyboard,
                             var=u'flush_keyboard')
        self.edit_vbox.addWidget(self.checkbox_flush_keyboard)
        self.treewidget = TreeOverview(self.main_window, overview_mode=False)
        self.treewidget.setup(self.main_window)
        self.treewidget.structure_change.connect(self.update)
        self.treewidget.text_change.connect(self.update_script)
        self.set_focus_widget(self.treewidget)
        self.edit_vbox.addWidget(self.treewidget)
        self.add_text(
            _(u'<b>Important</b>: A sequence has <a href="http://osdoc.cogsci.nl/usage/prepare-run">a variable preparation time</a>.'))

    def edit_widget(self):
        super().edit_widget()
        if self.treewidget.locked:
            return
        for item, cond, enabled in self.items:
            if item not in self.experiment.items:
                self.extension_manager.fire(
                    u'notify',
                    message=_(u'Sequence contains non-existing item: %s') % item,
                    category=u'warning')
        self.treewidget.clear()
        self.toplevel_treeitem = self.build_item_tree(max_depth=2)
        self.treewidget.addTopLevelItem(self.toplevel_treeitem)
        self.toplevel_treeitem.setExpanded(True)
        self.treewidget.resizeColumnToContents(0)
        self.treewidget.append_button.set_position()

    @requires_init
    @QtStructureItem.clears_children_cache
    def rename(self, from_name, to_name):
        QtPlugin.rename(self, from_name, to_name)
        new_items = []
        for item, cond, enabled in self.items:
            if item == from_name:
                new_items.append((to_name, cond, enabled))
            else:
                new_items.append((item, cond, enabled))
        self.items = new_items
        self.treewidget.rename(from_name, to_name)

    @QtStructureItem.clears_children_cache
    def delete(self, item_name, item_parent=None, index=None):
        if item_parent is None or (item_parent == self.name and index is None):
            while True:
                for i, (child_item_name, child_run_if) in enumerate(self.items):
                    if child_item_name == item_name:
                        self.items = self.items[:i]+self.items[i+1:]
                        break
                else:
                    # Break the while loop if no break occurred in the for loop
                    break
        elif item_parent == self.name and index is not None:
            if self.items[index][0] == item_name:
                self.items = self.items[:index]+self.items[index+1:]

    def build_item_tree(self, toplevel=None, items=[], max_depth=-1,
                        extra_info=None):
        widget = TreeItemItem(self, extra_info=extra_info)
        items.append(self.name)
        if max_depth < 0 or max_depth > 1:
            for item, cond, enabled in self.items:
                if item in self.experiment.items:
                    tree_widget = self.experiment.items[item].build_item_tree(
                        widget, items, max_depth=max_depth-1, extra_info=cond)
                    if not enabled:
                        tree_widget.setDisabled(True)
                        tree_widget.set_extra_info(None)
                    
        if toplevel is not None:
            toplevel.addChild(widget)
        else:
            widget.set_draggable(False)
        return widget

    def set_run_if(self, index, cond=u'always'):
        r"""Sets the run-if statement for an item at a specific index.

        Parameters
        ----------
        index : int
            The index of the item to change the run-if statement of.
        cond : unicode, optional
            The run-if statement.
        """
        self.items[index] = self.items[index][0], cond, self.items[index][2]

    @QtStructureItem.cached_children
    def children(self):
        """See qtitem."""
        self._children = []
        for item, _, _ in self.items:
            if item not in self.experiment.items:
                continue
            self._children += [item] + self.experiment.items[item].children()
        return self._children

    def direct_children(self):
        return [item for item, _, _ in self.items
                if item in self.experiment.items]

    def is_child_item(self, item):
        return item in self.children()

    @QtStructureItem.clears_children_cache
    def insert_child_item(self, item_name, index=0):
        if item_name == self.last_removed_child[0]:
            # If this item was just removed, re-add it and preserve its run-if
            # statement.
            self.items.insert(index, self.last_removed_child)
        else:
            self.items.insert(index, (item_name, 'True', True))
        self.update()
        self.main_window.set_unsaved(True)

    @QtStructureItem.clears_children_cache
    def remove_child_item(self, item_name, index=0):
        if index < 0:
            items = []
            for item, cond, enabled in self.items:
                if item != item_name:
                    items.append((item, cond, enabled))
            self.items = items
        elif len(self.items) > index and self.items[index][0] == item_name:
            # We remember the last removed child item, because we will reuse
            # it's run-if statement if it is re-added.
            self.last_removed_child = self.items[index]
            del self.items[index]
        if not self.update():
            self.extension_manager.fire(u'change_item', name=self.name)
        self.main_window.set_unsaved(True)

    def enable_child_item(self, item_name, index=0, enabled=True):
        # If no index is specified, we set the disabled status of all children
        if index < 0:
            items = []
            for item, cond, _enabled in self.items:
                if item == item_name:
                    _enabled = enabled
                items.append((item, cond, _enabled))
            self.items = items
        elif len(self.items) > index and self.items[index][0] == item_name:
            self.items[index] = item_name, self.items[index][1], enabled
        if not self.update():
            self.extension_manager.fire(u'change_item', name=self.name)
        self.main_window.set_unsaved(True)


# Alias for backwards compatibility
sequence = Sequence



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
import functools
import sys
import io
import yaml

py3 = True


def safe_decode(s, enc='utf-8', errors='strict'):
    if isinstance(s, str):
        return s
    if isinstance(s, bytes):
        return s.decode(enc, errors)
    # Numeric values are encoded right away
    if isinstance(s, (int, float)):
        return str(s)
    # Some types need to be converted to unicode, but require the encoding
    # and errors parameters. Notable examples are Exceptions, which have
    # strange characters under some locales, such as French. It even appears
    # that, at least in some cases, they have to be encodeed to str first.
    # Presumably, there is a better way to do this, but for now this at
    # least gives sensible results.
    if isinstance(s, Exception):
        try:
            return safe_decode(bytes(s), enc=enc, errors=errors)
        except:
            pass
    # For other types, the unicode representation doesn't require a specific
    # encoding. This mostly applies to non-stringy things, such as integers.
    try:
        return str(s)
    except Exception as e:
        from libopensesame.oslogging import oslogger
        oslogger.error(u'%s occurred, returning empty string' % type(e))
        return u''


def safe_encode(s, enc='utf-8', errors='strict'):
    if isinstance(s, bytes):
        return s
    if isinstance(s, str):
        return s.encode(enc, errors)
    try:
        return bytes(s)
    except Exception as e:
        from libopensesame.oslogging import oslogger
        oslogger.error(u'%s occurred, returning empty string' % type(e))
        return b''


def safe_read(path):
    with safe_open(path, u'r') as fd:
        try:
            s = fd.read()
        except UnicodeDecodeError:
            return u''
    return safe_decode(s, errors=u'ignore')


def safe_yaml_load(s):
    return yaml.load(s, Loader=yaml.UnsafeLoader)


safe_str = safe_decode
safe_open = functools.partial(open, encoding=u'utf-8')
basestring = str  # deprecated

__all__ = [
    'py3',
    'safe_decode',
    'safe_encode',
    'safe_str',
    'safe_read',
    'safe_open',
    'safe_yaml_load',
    'basestring'  # deprecated
]



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp import backend


class ElementFactory:

    def __init__(self, *args, **kwargs):

        self._args = args
        self._kwargs = kwargs

    def construct(self, canvas):

        bck = backend.backend_guess(canvas.experiment, u'canvas')
        mod = __import__('openexp._canvas._%s.%s' % (self.mod, bck),
                         fromlist=['dummy'])
        cls = getattr(mod, bck.capitalize())
        return cls(canvas, *self._args, **self._kwargs)


class Line(ElementFactory):
    mod = 'line'


class Rect(ElementFactory):
    mod = 'rect'


class Ellipse(ElementFactory):
    mod = 'ellipse'


class Circle(ElementFactory):
    mod = 'circle'


class FixDot(ElementFactory):
    mod = 'fixdot'


class Polygon(ElementFactory):
    mod = 'polygon'


class Image(ElementFactory):
    mod = 'image'


class Gabor(ElementFactory):
    mod = 'gabor'


class NoisePatch(ElementFactory):
    mod = 'noise_patch'


class RichText(ElementFactory):
    mod = 'richtext'


class Arrow(ElementFactory):
    mod = 'arrow'


Text = RichText



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.exceptions import InvalidValue


def float_list(l, desc, min_len=None, max_len=None):
    """Converts a variable to a list of floats if possible.

    Parameters
    ----------
    l
        The variable to convert.
    desc
        A description to clarify the osexception.
    min_len, optional
        The minimum length of the list.
    max_len, optional
        The maximum length of the list.

    Raises
    ------
    InvalidValue
        If the variable could not be converted.

    Returns
    -------
    list of float
    """
    try:
        l = list(l)
    except:
        raise InvalidValue(
            f'Expecting a list or similar type not {l} for {desc}')
    if min_len is not None and len(l) < min_len:
        raise InvalidValue(
            f'Expecting a list of at least {min_len} items for {desc}')
    if max_len is not None and len(l) > max_len:
        raise InvalidValue(
            f'Expecting a list of at most {max_len} items for {desc}')
    return l



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import copy
from libopensesame.exceptions import InvalidValue
from functools import partial
from openexp.color import Color

INF = float('inf')
NUMERIC_PROPERTIES = 'x', 'y', 'w', 'h', 'r'
COLOR_PROPERTIES = 'color', 'col1', 'col2'


class Element:
    """A base class for sketchpad elements.

    Parameters
    ----------
    canvas : Canvas
        The canvas of which this element is part.
    **properties : dict
        A dict with style arguments such as color, fill, etc.
    """
    # A property that indicates whether style properties (color etc) can be
    # changed or not.
    read_only = False

    def __init__(self, canvas, **properties):
        self._canvas = canvas
        if 'visible' not in properties:
            properties['visible'] = True
        for prop in COLOR_PROPERTIES:
            if prop in properties:
                properties[prop] = Color(self.experiment, properties[prop])
        self._assert_numeric(**{
            prop: val
            for prop, val in properties.items()
            if prop in NUMERIC_PROPERTIES
        })
        self._properties = properties
        for prop in self.property_names:
            self._create_property(prop)
        if canvas.auto_prepare and self.visible:
            self.prepare()

    def __contains__(self, xy):

        x, y, w, h = self.rect
        return x <= xy[0] and x+w >= xy[0] and y <= xy[1] and y+h >= xy[1]

    def __iter__(self):
        """Elements are iterable, but by default contain only themselves.
        However, Group objects can contain other elements.
        """
        yield self

    def __len__(self):
        """Elements have a length, but by default this is 1. However, Group
        objects can have a different length.
        """
        return 1

    def __add__(self, element):
        """Implements the + syntax, which combines Element objects into Group
        objects.

        Parameters
        ----------
        element : Element
            The element to add.

        Returns
        -------
        Group
            A group of elements.
        """
        from openexp._canvas._element.group import Group
        return Group(self.canvas, [self, element])

    def copy(self, canvas):
        """Creates a deep copy of the current element. This new copy becomes
        part of the provided canvas.

        Parameters
        ----------
        canvas : Canvas
            The canvas of which the copied element is part.

        Returns
        -------
        Element
            A copy of the current element.
        """
        try:
            e = copy.deepcopy(self)
        except (ValueError, TypeError):
            # Silently fall back to shallow copies if deep copies are not
            # possible. This happens for example with the c-types-based
            # PsychoPy stimuli.
            e = copy.copy(self)
        e._canvas = canvas
        return e

    def prepare(self):
        """Is called when the canvas is prepared. This should be implemented
        by backend-specific element objects.
        """
        pass

    def show(self):
        """Is called when the canvas is shown. This should be implemented by
        backend-specfic element objects.
        """
        pass

    def _on_attribute_change(self, **kwargs):
        """Is called when an attribute, such as color, is changed.

        Parameters
        ----------
        **kwargs : dict
            A dict with changed attributes.
        """
        pass

    @property
    def experiment(self):
        return self._canvas.experiment

    @property
    def to_xy(self):
        return self._canvas.to_xy

    @property
    def none_to_center(self):
        return self._canvas.none_to_center

    @property
    def property_names(self):
        return (
            set(self._properties.keys())
            | set(self._canvas.configurables.keys())
        )

    @property
    def rect(self):
        raise NotImplementedError()

    @property
    def top(self):
        return self.rect[1]

    @property
    def left(self):
        return self.rect[0]

    @property
    def width(self):
        return self.rect[2]

    @property
    def height(self):
        return self.rect[3]

    @property
    def size(self):
        return self.rect[2:]

    @property
    def position(self):
        return self.rect[:2]

    def _create_property(self, key):
        """Dynamically creates a getter/setter property. This is used to make
        style arguments such as color get-able and set-able.
        """
        setattr(self.__class__, key, property(
                partial(self._getter, key),
                partial(self._setter, key),
                self._deller, ''))

    @staticmethod
    def _getter(key, self):
        """A getter for dynamically created properties.

        Parameters
        ----------
        key : str
            A property name.
        self : Element.
            The Element instance. For technical reasons this is passed as the
            second argument.
        """
        try:
            return self._properties[key]
        except KeyError:
            return self._canvas.__cfg__[key]

    @staticmethod
    def _setter(key, self, val):
        """A setter for dynamically created properties.

        Parameters
        ----------
        key : str
            A property name.
        self : Element.
            The Element instance. For technical reasons this is passed as the
            second argument.
        val : Any
            A property value.
        """
        if key in NUMERIC_PROPERTIES:
            self._assert_numeric(**{key: val})
        if key == 'color':
            val = color(self.experiment, val)
        self._properties[key] = val
        self._on_attribute_change(**{key: val})

    @staticmethod
    def _deller(self, key):
        """A deller for dynamically created properties.

        Parameters
        ----------
        key : str
            A property name.
        """
        pass

    @staticmethod
    def _assert_numeric(**kwdict):

        for name, v in kwdict.items():
            try:
                v = float(v)
            except ValueError:
                raise InvalidValue(
                    '%s should be int or float, not %s' % (name, v))
            if v != v:
                raise InvalidValue(
                '%s should be int or float, not nan' % name)
            if v == INF:
                raise InvalidValue(
                    '%s should be int or float, not inf' % name)

    @staticmethod
    def _rect(x, y, w, h):
        """Fixes negative width and heights when defining a rect

        Parameters
        ----------
        x
            An X coordinate
        y
            A Y coordinate
        w
            A width
        h
            A height

        Returns
        -------
        """
        Element._assert_numeric(x=x, y=y, w=w, h=h)
        if w < 0:
            x += w
            w = abs(w)
        if h < 0:
            y += h
            h = abs(h)
        return x, y, w, h



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *


class OSExpBase:

    def __init__(self, exp):

        self._format = None
        self._exp = exp

    @staticmethod
    def valid_extension(path):
        r"""Checks whether a path ends with one of the known extensions for an
        experiment file.

        Parameters
        ----------
        path
            The path to check.

        Returns
        -------
        bool
        """
        return any(
            path.lower().endswith(ext) for ext in [
                u'.opensesame',
                u'.opensesame.tar.gz',
                u'.osexp'
            ]
        )

    @property
    def format(self):
        r"""The format of the experiment, which can be:

        - script: a script
        (ie. not from file)
        - scriptfile: plain-text file
        - targz: targz
        archive
        - tar: tar archive
        """
        if self._format is None:
            self._format = self._determine_format()
        return self._format

    @property
    def experiment_path(self):
        r"""The path to the experiment file, or None if the experiment was
        loaded from script.
        """
        return self._experiment_path

    @property
    def script(self):
        r"""The experiment script."""
        raise NotImplementedError()

    def _determine_format(self):

        raise NotImplementedError()

    @property
    def _syntax(self):
        r"""A shortcut to the experiment syntax object."""
        return self._exp._syntax

    @property
    def _pool(self):
        r"""A shortcut to the experiment pool object."""
        return self._exp.pool



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.oslogging import oslogger
from libopensesame.base_response_item import BaseResponseItem
from openexp.keyboard import Keyboard
from .libsrbox import LibSrbox


class Srbox(BaseResponseItem):

    process_feedback = True

    def reset(self):
        self.var.timeout = u'infinite'
        self.var.lights = u''
        self.var.dev = u'autodetect'
        self.var._dummy = u'no'
        self.var.require_state_change = u'no'

    def validate_response(self, response):
        try:
            response = int(response)
        except (ValueError, TypeError):
            return False
        return 0 <= response <= 255

    def process_response(self, response_args):
        response, t1 = response_args
        if isinstance(response, list):
            response = response[0]
        super().process_response((safe_decode(response), t1))

    def _get_button_press(self):
        r"""Calls srbox.get_button_press() with the correct arguments."""
        return self.experiment.srbox.get_button_press(
            allowed_buttons=self._allowed_responses,
            timeout=self._timeout,
            require_state_change=self._require_state_change
        )

    def prepare_response_func(self):
        self._keyboard = Keyboard(self.experiment,
                                  keylist=self._allowed_responses,
                                  timeout=self._timeout)
        if self.var._dummy == u'yes':
            return self._keyboard.get_key
        # Prepare the device string
        dev = self.var.dev
        if dev == u"autodetect":
            dev = None
        # Dynamically create an srbox instance
        if not hasattr(self.experiment, "srbox"):
            self.experiment.srbox = LibSrbox(self.experiment, dev)
            self.experiment.cleanup_functions.append(self.close)
            self.python_workspace[u'srbox'] = self.experiment.srbox
        # Prepare the light byte
        s = "010"  # Control string
        for i in range(5):
            if str(5 - i) in str(self.var.lights):
                s += "1"
            else:
                s += "0"
        self._lights = chr(int(s, 2))
        oslogger.debug(u"lights string set to %s (%s)" % (s, self.var.lights))
        if self._allowed_responses is not None:
            self._allowed_responses = [int(r) for r in self._allowed_responses]
        self._require_state_change = self.var.require_state_change == u'yes'
        return self._get_button_press

    def run(self):
        """See item."""
        self._keyboard.flush()
        if self.var._dummy != u'yes':
            self.experiment.srbox.start()
        super().run()
        if self.var._dummy != u'yes':
            self.experiment.srbox.stop()

    def close(self):
        r"""Neatly close the connection to the srbox."""
        if (
                not hasattr(self.experiment, "srbox") or
                self.experiment.srbox is None
        ):
            oslogger.debug("no active srbox")
            return
        try:
            self.experiment.srbox.close()
            self.experiment.srbox = None
            oslogger.debug("srbox closed")
        except Exception as e:
            oslogger.error("failed to close srbox")



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
from openexp import resources
from libopensesame.widgets.themes.gray import Gray
from libopensesame.widgets.themes.plain import Plain


class GrayLarge(gray):

    def __init__(self, form):

        Plain.__init__(self, form)
        self.box_checked_image = \
            resources['widgets/gray/box-checked-large.png']
        self.box_unchecked_image = \
            resources['widgets/gray/box-unchecked-large.png']

    def box_size(self):

        return 32


# Alias for backwards compatibility
gray_large = GrayLarge



# coding=utf-8

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libqtopensesame.extensions import BaseExtension
from .tab_to_dockwidget_docktab import DockTab
from qtpy import QtCore


class TabToDockwidget(BaseExtension):

    r"""This extensions turns tabs into standalone dockwidgets, so that you can
    see multiple things side by side.
    """
    def event_startup(self):
        r"""During initialization we monkey-patch the tabwidget so that it also
        manages dockwidgets.
        """
        self._docktabs = {}
        self.overview_area.structure_change.connect(self.update_all)
        self.tabwidget.add = self._add_tab(self.tabwidget.add)
        self.tabwidget.remove = self._remove_tab(self.tabwidget.remove)
        self.tabwidget.close_all = self._close_all_tabs(
            self.tabwidget.close_all)

    def event_run_experiment(self, fullscreen):
        r"""Disable all docked item controls when running the experiment."""
        for docktab in self._docktabs.values():
            docktab.setEnabled(False)

    def event_end_experiment(self, ret_val):
        r"""Re-enable all docked item controls when the experiment is done."""
        for docktab in self._docktabs.values():
            docktab.setEnabled(True)

    def update_all(self):
        r"""Update all item controls."""
        for docktab in self._docktabs.values():
            docktab.update()

    def event_rename_item(self, from_name, to_name):
        r"""Change the dockwidget titles on item rename, and allow items to
        update in response to renames.
        """
        for docktab in self._docktabs.values():
            docktab.rename(from_name, to_name)

    def event_new_item(self, name, _type):
        r"""Allow items to update in response to new items."""
        self.update_all()

    def event_delete_item(self, name):
        r"""Allow items to update in response to item deletions."""
        for docktab in self._docktabs.values():
            docktab.delete(name)

    def provide_item_in_dockwidget(self, name):
        r"""Checks whether an item (by name) is currently shown as a
        dockwidget.
        """
        if name not in self.item_store:
            return False
        return self.item_store[name].widget() in self._docktabs

    def activate(self):
        r"""Turns the currently visible tab (if any) into a dockwidget."""
        widget = self.tabwidget.currentWidget()
        if widget is None:
            return
        name = self.tabwidget.tabText(self.tabwidget.currentIndex())
        self.tabwidget.remove(widget)
        docktab = DockTab(self, widget, name)
        self._docktabs[widget] = docktab
        self.main_window.addDockWidget(QtCore.Qt.RightDockWidgetArea, docktab)

    def remove_widget(self, widget):
        r"""Removes a docked widget."""
        if widget in self._docktabs:
            del self._docktabs[widget]

    def _close_all_tabs(self, fnc):
        r"""Decorates tabwidget.close_all()"""
        def inner(*args, **kwargs):

            while self._docktabs:
                widget, docktab = self._docktabs.popitem()
                docktab.close()
            return fnc(*args, **kwargs)

        return inner

    def _add_tab(self, fnc):
        r"""Decorates tabwidget.add()"""
        def inner(widget, *args, **kwargs):

            if widget in self._docktabs:
                if hasattr(widget, u'set_focus'):
                    widget.set_focus()
                return
            return fnc(widget, *args, **kwargs)

        return inner

    def _remove_tab(self, fnc):
        r"""Decorates tabwidget.remove()"""
        def inner(widget, *args, **kwargs):

            if widget in self._docktabs:
                docktab = self._docktabs.pop(widget)
                docktab.close()
                return
            return fnc(widget, *args, **kwargs)

        return inner



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.item import Item
from libopensesame.exceptions import InvalidValue


class BaseResponseItem(Item):
    r"""The base class for items that collect responses, such as
    keyboard_response, mouse_response, joystick, etc.
    """
    
    # Override as True for objects that should be included in feedback
    process_feedback = False

    def prepare_response_func(self):
        r"""Should return a function that, when called, returns a (response,
        timestamp) tuple. This function needs to be implemented in every
        response item.

        Returns
        -------
        FunctionType
        """
        raise NotImplementedError()

    def validate_response(self, response):
        r"""Optionally checks whether a response is valid for this item. This
        can be used to check whether the list of allowed responses is valid.

        Parameters
        ----------
        response
            The response to check.

        Returns
        -------
        bool
            True if response is valid, False otherwise.
        """
        return True

    def response_matches(self, test, ref):
        r"""Checks whether two responses are the same. This can be re-
        implemented to take synonyms into account.

        Parameters
        ----------
        test
            The first response.
        ref
            The second response.

        Returns
        -------
        bool
            True if the responses match, False otherwise.
        """
        return test in ref

    def process_response(self, response_args):
        r"""Processes a collected responses, so that it is added to the
        response_store, etc.

        Parameters
        ----------
        response_args : tuple
            The return value of the function created by
            prepare_response_func().
        """
        response, t1 = response_args
        if self._correct_responses is not None:
            correct = self.response_matches(response, self._correct_responses)
        else:
            correct = None
        self.responses.add(
            response=response, response_time=t1-self._t0, correct=correct,
            item=self.name, feedback=self.process_feedback)

    def prepare(self):
        super().prepare()
        self._timeout = self._prepare_timeout()
        self._allowed_responses = self._prepare_responses('allowed_responses')
        self._correct_responses = self._prepare_responses('correct_response')
        self._collect_response = self.prepare_response_func()
        self._t0 = None

    def run(self):
        super().run()
        if self._t0 is None:
            self._t0 = self.set_item_onset()
        retval = self._collect_response()
        if retval is None:
            return
        self.process_response(retval)

    def var_info(self):
        l = []
        l.append(("response", "[Depends on response]"))
        l.append(("response_time", "[Depends on response]"))
        l.append(("response_%s" % self.name, "[Depends on response]"))
        l.append(("response_time_%s" % self.name, "[Depends on response]"))
        if self.process_feedback:
            l.append(("correct", "[Depends on response]"))
            l.append(("correct_%s" % self.name, "[Depends on response]"))
            l.append(("average_response_time", "[Depends on response]"))
            l.append(("avg_rt", "[Depends on response]"))
            l.append(("accuracy", "[Depends on response]"))
            l.append(("acc", "[Depends on response]"))
        return super().var_info() + l

    # Private functions

    def _prepare_timeout(self):
        r"""Processes the timeout variable, and checks whether it is valid.

        Returns
        -------
        float
            A timeout value.
        """
        timeout = self.var.get('timeout', default='infinite')
        if timeout == 'infinite':
            return
        try:
            timeout = int(timeout)
            assert(timeout >= 0)
        except Exception:
            raise InvalidValue(f'"{timeout}" is not a valid timeout. '
                               f'Expecting a positive integer or "infinite".')
        return timeout

    def _prepare_responses(self, var):
        r"""Parses a semicolon-separated list of responses into a list of
        responses or None if no responses are provided.

        Parameters
        ----------
        var : str
            The name of the variable that contains the responses.

        Returns
        -------
        NoneType, list
        """
        responses = safe_decode(self.var.get(var, default=''))
        if responses == '':
            return
        responses = [r.strip() for r in responses.split(';')]
        for r in responses:
            if not self.validate_response(r):
                raise InvalidValue(f'Invalid value in {var}: {r}')
        # If allowed responses are provided, the list should not be empty
        if not responses:
            raise InvalidValue(f'{var} should not be an empty list')
        return responses

    def _prepare_sleep_func(self, duration):
        r"""Creates a function that sleeps for a specific duration.

        Parameters
        ----------
        duration : int, float
            The duration to sleep for.

        Returns
        -------
        FunctionType
            A sleep function with a fixed duration.
        """
        if duration == 0:
            return lambda: None
        if self.var.duration > 0:
            return lambda: self.clock.sleep(self.var.duration)
        raise InvalidValue('Duration should not be negative')


# Alias for backwards compatibility
base_response_item = BaseResponseItem



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from qtpy import QtWidgets
from libqtopensesame.widgets.base_widget import BaseWidget
from libqtopensesame.misc.translate import translation_context
from libqtopensesame.dialogs.text_input import TextInput
_ = translation_context(u'logger', category=u'item')


class LoggerWidget(BaseWidget):
    """The logger widget

    Parameters
    ----------
    logger: logger
        A logger object.
    """
    
    def __init__(self, logger):
        super().__init__(logger.main_window, ui=u'widgets.logger')
        self.logger = logger

    def update(self):
        self.ui.textedit_include.setPlainText('\n'.join(self.logger.logvars))
        self.ui.textedit_exclude.setPlainText(
            '\n'.join(self.logger.exclude_patterns))


# Alias for backwards compatibility
logger_widget = LoggerWidget


# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os.path
import sys
import imp
import qtawesome as qta
from openexp import resources
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.config import cfg
from qtpy import QtGui, QtWidgets, QtCore

available_themes = [u'default', u'monokai']


class Theme:

    """Handles the GUI theme"""
    default_icon_size = 32

    def __init__(self, main_window, theme=None):
        """
        Constructor

        Arguments:
        main_window -- the main_window object

        Keyword arguments:
        theme -- the theme to be used or None to use config (default=None)
        """
        self.main_window = main_window
        self.fallback_icon = QtGui.QIcon(resources['theme/fallback.png'])
        self.theme = cfg.theme if theme is None else theme
        self.theme_folder = resources[f'theme/{self.theme}']
        self._icon_theme_path = self.theme_folder
        oslogger.debug(u"theme = '%s' (%s)" % (self.theme, self.theme_folder))
        # The theme folder must exist, and contain a file called __theme__.py,
        # if not, we fall back to the default theme, which is assumed to always
        # exist.
        if self.theme_folder is None or not os.path.exists(
                os.path.join(self.theme_folder, u'__theme__.py')
        ):
            oslogger.warning(
                u"theme '%s' does not exist, using 'default'" % theme,
            )
            self.theme = u"default"
            self._icon_theme_path = self.theme_folder = \
                resources[f'theme/{self.theme}']
        self.theme_info = os.path.join(self.theme_folder, u"__theme__.py")
        if os.path.exists(self.theme_info):
            info = imp.load_source(
                self.theme,
                safe_str(self.theme_info, enc=sys.getfilesystemencoding())
            )
            with safe_open(os.path.join(self.theme_folder, info.qss)) as fd:
                self._qss = fd.read()
            self._icon_map = info.icon_map
            self._icon_theme = info.icon_theme
            if hasattr(info, 'qdatamatrix'):
                from qdatamatrix import _qcell
                for key, color in info.qdatamatrix.items():
                    setattr(_qcell, key, color)
            if hasattr(info, 'icon_theme_path'):
                self._icon_theme_path = \
                    resources[f'theme/{info.icon_theme_path}']
        self.load_icon_map()
        self.apply_theme(self.main_window)

    @property
    def experiment(self):
        return self.main_window.experiment

    def apply_theme(self, widget):
        """
        Apply the theme to a QWidget, i.e. load the stylesheet and the icons

        Arguments:
        widget -- a QWidget
        """
        if hasattr(widget, u'setStyleSheet'):
            widget.setStyleSheet(self._qss)
        if hasattr(widget, u"ui"):
            self.load_icons(widget.ui)

    def qfileicon(self, path):
        r"""Gets an filetype icon for a file.

        Parameters
        ----------
        path
            The full path to the file. The file must exist.

        Returns
        -------
        QIcon
        """
        try:
            import fileinspector
        except ImportError:
            return QtWidgets.QFileIconProvider().icon(QtCore.QFileInfo(path))
        filetype = fileinspector.determine_type(path, output='xdg')
        if filetype is None:
            filetype = u'text-x-generic'
        category = fileinspector.determine_category(filetype)
        if category is None:
            fallback = u'text-x-generic'
        else:
            fallback = category + u'-x-generic'
        return QtGui.QIcon.fromTheme(
            filetype,
            QtGui.QIcon.fromTheme(fallback)
        )

    def qicon(self, icon):
        """
        desc:
                Gets an icon from the theme.

        arguments:
                icon:	One of the following:
                                - A QIcon, which returned as is
                                - The name of an image file
                                - The name of a plug-in with a hardcoded icon
                                - The name of an entry in the icon map
                                - A theme icon name.

        returns:
                desc:	An icon, or a fallback icon if the specified wasn't found.
                type:	QIcon
        """
        if isinstance(icon, QtGui.QIcon):
            return icon
        if os.path.isfile(icon):
            qicon = QtGui.QIcon()
            if icon.endswith(u'_large.png'):
                size = 32
            else:
                size = 16
            qicon.addFile(icon, size=QtCore.QSize(size, size))
            return qicon
        if (
                hasattr(self, u'experiment') and
                (
                    u'%s_large.png' % icon in self.experiment.resources or
                    '%s.png' in self.experiment.resources
                )
        ):
            qicon = QtGui.QIcon()
            if u'%s_large.png' % icon in self.experiment.resources:
                qicon.addFile(resources[f'{icon}_large.png'],
                              size=QtCore.QSize(32, 32))
            if u'%s.png' % icon in self.experiment.resources:
                qicon.addFile(resources[f'{icon}.png'],
                              size=QtCore.QSize(16, 16))
            return qicon
        if icon in self.icon_map:
            name = self.icon_map[icon][0]
        else:
            name = icon
        icon = QtGui.QIcon.fromTheme(name, self.fallback_icon)
        if icon.name() != name:
            oslogger.debug(
                u'missing icon {}, using {}'.format(name, icon.name())
            )
        return icon

    def qpixmap(self, icon, size=None):
        """
        Get an icon from the theme

        Arguments:
        icon -- the icon name

        Keyword arguments:
        size -- the size of the icon or None for default (default=None)

        Returns:
        A QPixmap
        """
        if isinstance(icon, str) and os.path.exists(icon):
            return QtGui.QPixmap(icon)
        if size is None:
            if icon in self.icon_map:
                size = self.icon_map[icon][1]
            else:
                size = self.default_icon_size
        size = QtCore.QSize(size, size)
        # Pixmap should never return a pixmap larger than the requested size.
        # However, this does happen on some versions of Qt4. Therefore, we check
        # the size, and scale if necessary.
        pixmap = self.qicon(icon).pixmap(size)
        if pixmap.size() != size:
            pixmap = pixmap.scaled(size)
        return pixmap

    def qlabel(self, icon, size=None):
        """
        Get an icon from the theme

        Arguments:
        icon -- the icon name

        Keyword arguments:
        size -- the size of the icon or None for default (default=None)

        Returns:
        A QLabel
        """
        label = QtWidgets.QLabel()
        label.setPixmap(self.qpixmap(icon, size=size))
        return label

    def load_icon_map(self):
        """Load the icon map"""
        self.original_theme = QtGui.QIcon.themeName()
        if os.path.exists(
                os.path.join(self._icon_theme_path, self._icon_theme)
        ):
            oslogger.debug(u"using custom icon theme")
            QtGui.QIcon.setThemeSearchPaths(
                QtGui.QIcon.themeSearchPaths() + [self._icon_theme_path]
            )
            QtGui.QIcon.setThemeName(self._icon_theme)
        else:
            oslogger.warning(u"using default icon theme, icons may be missing")
        self.icon_map = {}
        path = os.path.join(self.theme_folder, self._icon_map)
        with safe_open(path) as fd:
            for line in fd:
                line = line.split(u',')
                if len(line) != 3:
                    continue
                try:
                    size = int(line[2])
                except ValueError:
                    size = 32
                alias = line[0].strip()
                name = line[1].strip()
                if alias in self.icon_map:
                    oslogger.debug(
                        u"alias '%s' already in icon map, overwriting" % alias,
                        reason=u"warning"
                    )
                self.icon_map[alias] = name, size

    def load_icons(self, ui):
        """
        Add icons to all icon-supporting widgets in a ui object. Icons can
        be specified through a `qta:[icon name]` text on the widget, or through
        the icon_map, where the object name is the look-up key.

        Parameters
        ----------
        ui
            The ui object to load icons into
        """
        if ui is None:
            return
        for key, value in ui.__dict__.items():
            if not isinstance(value, QtCore.QObject):
                continue
            if hasattr(value, 'text') and value.text().startswith('qta:'):
                try:
                    value.setIcon(qta.icon(value.text()[4:]))
                    value.setText('')
                except Exception as e:
                    oslogger.warning(f'failed to load qtawesome icon: {e}')
                    
            elif key in self.icon_map:
                if hasattr(value, 'setIcon'):
                    value.setIcon(self.qicon(key))
                elif hasattr(value, 'setPixmap'):
                    value.setPixmap(self.qpixmap(key))

    def set_toolbar_size(self, size):
        """
        Control the size of the icons in the toolbar

        Arguments:
        size -- a size in pixels
        """
        self.main_window.ui.toolbar_main.setIconSize(QtCore.QSize(size, size))
        self.main_window.ui.toolbar_items.setIconSize(QtCore.QSize(size, size))
        self.main_window.ui.toolbar_items.build()

    def resource(self, fname):
        """
        Retrieves the path to a resource within the theme folder.

        Arguments:
        fname	--	The resource filename.

        Returns:
        The full path to the resource file in the theme folder.
        """
        return os.path.join(self.theme_folder, fname)


# Alias for backwards compatibility
theme = Theme



"""A simple multiple choice item"""

category = "Form"
icon = "os-form_multiple_choice"
help_url = "manual/forms/readymade"
controls = [
    {
        "label": "Form title",
        "name": "line_edit_form_title",
        "tooltip": "Form title",
        "type": "line_edit",
        "var": "form_title"
    },
    {
        "label": "Response variable",
        "tooltip": "Response variable",
        "type": "line_edit",
        "var": "form_var"
    },
    {
        "label": "Allow multiple options to be selected",
        "name": "checkbox_allow_multiple",
        "tooltip": "Allow multiple options to be selected",
        "type": "checkbox",
        "var": "allow_multiple"
    },
    {
        "label": "Advance immediately to the next item once a selection has been made",
        "name": "checkbox_advance_immediately",
        "tooltip": "Advance immediately to the next item once a selection has been made",
        "type": "checkbox",
        "var": "advance_immediately"
    },
    {
        "label": "Button text",
        "name": "line_edit_button_text",
        "tooltip": "Text for the button to advance to the next item",
        "type": "line_edit",
        "var": "button_text"
    },
    {
        "label": "Timeout",
        "name": "line_edit_timeout",
        "tooltip": "Response timeout",
        "type": "line_edit",
        "validator": "timeout",
        "var": "timeout"
    },
    {
        "label": "Your question",
        "name": "editor_widget_question",
        "syntax": False,
        "tooltip": "Your question",
        "type": "editor",
        "var": "question"
    },
    {
        "label": "Response options (different options on different lines)",
        "name": "editor_widget_options",
        "syntax": False,
        "tooltip": "Response options",
        "type": "editor",
        "var": "options"
    }
]



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from qtpy import QtWidgets
from libqtopensesame.dialogs.patch_settings import PatchSettings
from libopensesame.py3compat import *


class GaborSettings(PatchSettings):

    r"""A gabor settings dialog."""
    def ui_file(self):
        """
        returns:
                desc:	The ui file.
                type:	unicode
        """
        return u'dialogs.gabor_settings'

    def get_properties(self):
        r"""Gets the Gabor properties.

        Returns
        -------
        dict
            A dictionary with properties.
        """
        if self.exec_() != QtWidgets.QDialog.Accepted:
            return None

        properties = self._spinbox_properties(
            u'orient',
            u'freq',
            u'size',
            u'stdev',
            u'phase'
        )
        properties.update(self._combobox_properties(
            (u'env', self.env_map),
            (u'bgmode', self.bgmode_map)
        ))
        properties[u'color1'] = self.ui.edit_color1.text()
        properties[u'color2'] = self.ui.edit_color2.text()
        return properties

    def set_properties(self, properties):
        r"""Fills the dialog controls based on a properties dictionary.

        Parameters
        ----------
        properties : dict
            A properties dictionary.
        """
        self._set_spinbox(self.ui.spinbox_orient, u'orient', properties)
        self._set_spinbox(self.ui.spinbox_size, u'size', properties)
        self._set_spinbox(self.ui.spinbox_stdev, u'stdev', properties)
        self._set_spinbox(self.ui.spinbox_freq, u'freq', properties)
        self._set_spinbox(self.ui.spinbox_phase, u'phase', properties)
        self.ui.edit_color1.setText(properties[u'color1'])
        self.ui.edit_color2.setText(properties[u'color2'])
        self._set_combobox(
            self.env_map,
            self.ui.combobox_env,
            u'env',
            properties
        )
        self._set_combobox(
            self.bgmode_map,
            self.ui.combobox_bgmode,
            u'bgmode',
            properties
        )


# Alias for backwards compatibility
gabor_settings = GaborSettings



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
import os
import sys
import mimetypes
from openexp import resources
from libopensesame.oslogging import oslogger
from qtpy import QtGui, QtCore, QtWidgets
from libqtopensesame.items.qtitem import QtItem
from libqtopensesame.widgets.color_edit import ColorEdit
from libopensesame import misc
from libopensesame.oslogging import oslogger
from libqtopensesame.misc.translate import translation_context
from libqtopensesame._input.conditional_expression import ConditionalExpression
_ = translation_context(u'qtplugin', category=u'core')


class QtPlugin(QtItem):
    """Provides basic functionality for plugin GUIs.
    
    Parameters
    ----------
    plugin_file
        The location of the plugin script file.
    """
    def __init__(self, plugin_file=None):
        if plugin_file is not None:
            # These lines makes sure that the icons and help file are
            # recognized
            self.plugin_folder = os.path.dirname(plugin_file)
            for ext in ['.html', '.md']:
                basename = self.item_type + ext
                path = os.path.join(self.plugin_folder, 'locale',
                                    self.main_window._locale, basename)
                if os.path.exists(path):
                    resources[basename] = path
                    continue
                path = os.path.join(self.plugin_folder, basename)
                if os.path.exists(path):
                    resources[basename] = path
            # Install a translation file if there is one. Most plugins have
            # their translations as part of the OpenSesame main translations.
            # However, separate plugins can bring their own translation.
            # We search both resources/locale and locale for qm translation
            # files.
            qm_path = os.path.join(
                self.plugin_folder, 'resources', 'locale',
                f'{self.main_window.locale}.qm')
            if not os.path.exists(qm_path):
                qm_path = os.path.join(
                    self.plugin_folder, 'locale',
                    f'{self.main_window.locale}.qm')
            if os.path.exists(qm_path):
                if self.main_window.translators:
                    oslogger.info(f'installing translator {qm_path}')
                    self._translator = self.main_window.translators.pop()
                    self._translator.load(qm_path)
                    QtCore.QCoreApplication.installTranslator(self._translator)
                else:
                    oslogger.warning('no translators objects available')
            self.init_item_icon()
        else:
            self.qicon = None
        self.lock = False
        super().__init__()

    def init_item_icon(self):

        # If the icon is not a file, then it corresponds to a theme icon
        icon = self.plugin_manager[self.item_type].icon
        if not os.path.isfile(icon):
            self.qicon = self.theme.qicon(icon)
            return
        # Otherwise it is specified through two pngs, one for small and one for
        # large
        icon_large = self.plugin_manager[self.item_type].icon_large
        self.qicon = QtGui.QIcon()
        self.qicon.addFile(icon, QtCore.QSize(16, 16))
        self.qicon.addFile(icon_large, QtCore.QSize(32, 32))

    def item_icon(self):
        """
        returns:
                desc:	The name of the item icon.
                type:	unicode
        """
        if self.qicon is not None:
            return self.qicon
        return self.item_type

    def edit_widget(self):
        r"""Updates the GUI controls."""
        super().edit_widget()
        self.auto_edit_widget()

    def add_control(self, label, widget, tooltip=None, min_width=200,
                    info=None):
        r"""Adds a generic control QWidget.

        Parameters
        ----------
        label
            A text label.
        widget
            A QWidget.
        tooltip, optional
            A tooltip text.
        min_width, optional
            A minimum width for the widget.
        info, optional
            Additional info, to be presented to the right of the control.
        """
        if tooltip is not None:
            try:
                widget.setToolTip(tooltip)
            except:
                pass
        if isinstance(min_width, int):
            widget.setMinimumWidth(min_width)
        row = self.edit_grid.rowCount()
        label = QtWidgets.QLabel(label)
        label.setAlignment(self.edit_grid.labelAlignment())
        if info is None:
            self.edit_grid.insertRow(row, label, widget)
        else:
            hbox = QtWidgets.QHBoxLayout()
            hbox.setContentsMargins(0, 0, 0, 0)
            hbox.setSpacing(12)
            hbox.addWidget(widget)
            info = QtWidgets.QLabel(info)
            info.setObjectName(u'control-info')
            hbox.addWidget(info)
            container = QtWidgets.QWidget()
            container.setLayout(hbox)
            self.edit_grid.insertRow(row, label, container)
        self.set_focus_widget(widget)

    def add_line_edit_control(self, var, label, validator=None, **kwdict):
        r"""Adds a QLineEdit control that is linked to a variable.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        validator, optional
            A validator object.
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        A QLineEdit widget.
        """
        edit = QtWidgets.QLineEdit()
        edit.editingFinished.connect(self.apply_edit_changes)
        edit.textEdited.connect(self.set_dirty)
        edit.setMinimumWidth(200)
        if validator is not None:
            edit.setValidator(validator)
        self.add_control(label, edit, **kwdict)
        if var is not None:
            self.auto_line_edit[var] = edit
        return edit
    
    def add_conditional_expression_control(self, var, label, verb, **kwdict):
        """Adds a line-edit control that is linked to a variable and is
        customized for conditional expressions.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        verb
            A verb that describes the expression, such as 'repeat' for a
            repeat-if expression.
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        ConditionalExpression
        """
        edit = ConditionalExpression(self.main_window)
        edit.verb = verb
        edit.editingFinished.connect(self.apply_edit_changes)
        edit.textEdited.connect(self.set_dirty)
        edit.setMinimumWidth(200)
        self.add_control(label, edit, **kwdict)
        if var is not None:
            self.auto_line_edit[var] = edit
        return edit

    def add_checkbox_control(self, var, label, **kwdict):
        r"""Adds a QCheckBox control that is linked to a variable.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        A QCheckBox widget.
        """
        checkbox = QtWidgets.QCheckBox(label)
        checkbox.clicked.connect(self.apply_edit_changes)
        self.add_control(u'', checkbox, **kwdict)
        if var is not None:
            self.auto_checkbox[var] = checkbox
        return checkbox

    def add_color_edit_control(self, var, label, **kwdict):
        r"""Adds a color_edit control.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        A color_edit widget.
        """
        edit = ColorEdit(self.main_window)
        edit.setMinimumWidth(200)
        edit.initialize(self.experiment)
        edit.textEdited.connect(self.apply_edit_changes)
        self.add_control(label, edit, **kwdict)
        if var is not None:
            self.auto_line_edit[var] = edit
        return edit

    def add_combobox_control(self, var, label, options, **kwdict):
        r"""Adds a QComboBox control that is linked to a variable.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        options
            A list of options.
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        A QComboBox widget.
        """
        combobox = QtWidgets.QComboBox()
        combobox.setMinimumWidth(200)
        for o in options:
            combobox.addItem(o)
        combobox.activated.connect(self.apply_edit_changes)
        self.add_control(label, combobox, **kwdict)
        if var is not None:
            self.auto_combobox[var] = combobox
        return combobox

    def add_doublespinbox_control(self, *args, **kwdict):
        r"""Adds a QDoubleSpinBox. See _add_spinbox_control()."""
        return self._add_spinbox_control(QtWidgets.QDoubleSpinBox, *args,
                                         **kwdict)

    def add_spinbox_control(self, *args, **kwdict):
        r"""Adds a QSpinBox. See _add_spinbox_control()."""
        return self._add_spinbox_control(QtWidgets.QSpinBox, *args, **kwdict)

    def _add_spinbox_control(self, cls, var, label, min_val, max_val,
                             prefix=u'', suffix=u'', **kwdict):
        r"""Adds a QSpinBox or QDoubleSpinBox control that is linked to a
        variable.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        min_val
            A minimum value.
        max_val
            A maximum value.
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        A QSpinBox widget.
        """
        spinbox = cls()
        spinbox.setMinimum(min_val)
        spinbox.setMaximum(max_val)
        spinbox.editingFinished.connect(self.apply_edit_changes)
        spinbox.valueChanged.connect(self.set_dirty)
        spinbox.setMinimumWidth(200)
        if prefix != u'':
            spinbox.setPrefix(prefix)
        if suffix != u'':
            spinbox.setSuffix(suffix)
        self.add_control(label, spinbox, **kwdict)
        if var is not None:
            self.auto_spinbox[var] = spinbox
        return spinbox

    def add_slider_control(self, var, label, min_val, max_val, left_label=u'',
                           right_label=u'', **kwdict):
        """
        Adds a QSlider control that is linked to a variable.

        arguments:
                var:		The associated variable.
                label:		The label text.
                min_val:	A minimum value.
                max_val:	A maximum value.

        Keyword arguments:
                left_label:		A label for the left side.
                right_label:	A label for the right side.

        keyword-dict:
                kwdict:		A keyword dict to be passed on to add_control().

        returns:
                A QSlider widget.
        """
        slider = QtWidgets.QSlider(QtCore.Qt.Horizontal)
        slider.setFocusPolicy(QtCore.Qt.NoFocus)
        slider.setGeometry(30, 40, 100, 30)
        slider.setRange(min_val, max_val)
        slider.setSingleStep(1000)
        # Take care of layout
        layout = QtWidgets.QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)
        if left_label:
            llabel = QtWidgets.QLabel()
            llabel.setText(left_label)
            layout.addWidget(llabel)
        layout.addWidget(slider)
        if right_label:
            rlabel = QtWidgets.QLabel()
            rlabel.setText(right_label)
            layout.addWidget(rlabel)
        slider.sliderMoved.connect(self.apply_edit_changes)
        if var is not None:
            self.auto_slider[var] = slider
        widget = QtWidgets.QWidget()
        widget.setLayout(layout)
        self.add_control(label, widget, **kwdict)
        return slider

    def add_filepool_control(self, var, label, **kwdict):
        r"""Adds a control to select a file from the file pool, and is linked
        to a variable.

        Parameters
        ----------
        var
            The associated variable.
        label
            The label text.
        click_func, optional
            A custom function to be called when a file is selected. If no
            click_func is specified, file selection will be handled
            automatically. (default=None)
        **kwdict : dict
            A keyword dict to be passed on to add_control().


        Returns
        -------
        A pool_select widget that contains the path of the selected file.
        """
        from libqtopensesame._input.pool_select import pool_select
        edit = pool_select(self.main_window)
        edit.editingFinished.connect(self.apply_edit_changes)
        if var is not None:
            self.auto_line_edit[var] = edit
        self.add_control(label, edit, **kwdict)
        return edit

    def add_editor_control(self, var, label, syntax=False, language='python'):
        """
        Adds an editor that is linked to a variable.

        arguments:
                var:		The associated variable.
                label:		The label text.

        keywords:
                syntax:		A boolean indicating whether Python syntax highlighting
                                        should be activated.
                language:	The name of a programming language

        returns:
                An editor widget.
        """
        if syntax:
            if language == 'python':
                from libqtopensesame.pyqode_extras.widgets import \
                    PythonCodeEdit as CodeEdit
            else:
                from libqtopensesame.pyqode_extras.widgets import \
                    FallbackCodeEdit as CodeEdit
        else:
            from libqtopensesame.pyqode_extras.widgets import \
                TextCodeEdit as CodeEdit
        editor = CodeEdit(self.main_window)
        if syntax and language != 'python':
            editor.setPlainText(
                '',
                mime_type=mimetypes.guess_type('basename.' + language)
            )
        editor.focusOutEvent = self._editor_focus_out
        if var is not None:
            self.auto_editor[var] = editor
        # This is a horrible hack to avoid the situation in which newly created
        # editors are closed on the open_experiment event, which should only
        # close editors linked to the old experiment. This only affects editors
        # that bypass the lazy-init system, and for these editors the close()
        # function should be bypassed exactly once.
        if not self.lazy_init:
            def ignore_close_once(editor):
                def inner():
                    oslogger.debug(
                        'ignoring close() once for {}'.format(editor)
                    )
                    editor.close = orig_close
                orig_close = editor.close
                return inner
            editor.close = ignore_close_once(editor)
        self.edit_vbox.addWidget(editor)
        self.set_focus_widget(editor)
        self.extension_manager.fire(u'register_editor', editor=editor)
        return editor

    def _editor_focus_out(self, event):

        self.apply_edit_changes()

    def add_text(self, msg):
        r"""Adds a non-interactive QLabel for description purposes.

        Parameters
        ----------
        msg
            A text message.

        Returns
        -------
        A QLabel.
        """
        label = QtWidgets.QLabel(msg)
        label.setWordWrap(True)
        label.setOpenExternalLinks(True)
        self.edit_vbox.addWidget(label)
        return label

    def add_widget(self, w):
        r"""Adds a QWidget to the controls.

        Parameters
        ----------
        w : QWidget
            A widget.

        Returns
        -------
        QWidget
            The widget.
        """
        self.edit_vbox.addWidget(w)
        return w

    def add_stretch(self):
        r"""Pads empty space below the controls."""
        self.edit_vbox.addStretch()

    def get_ready(self):
        r"""Applies pending script changes.

        Returns
        -------
        True if changes have been made, False otherwise.
        """
        for var, editor in self.auto_editor.items():
            if editor.dirty:
                oslogger.debug(u'applying pending editor changes')
                self.apply_edit_changes()
                return True
        return super().get_ready()


# Alias for backwards compatibility
qtplugin = QtPlugin



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from openexp._log.log import Log
import os


class Csv(Log):

    r"""For docstrings, see openexp._log.log."""
    def __init__(self, experiment, path):

        self._log = None
        Log.__init__(self, experiment, path)

    def close(self):

        if self._log is not None:
            self._log.close()

    def open(self, path):

        if self._log is not None:
            self.close()
        # If only a filename is present, we interpret this filename as relative
        # to the experiment folder, instead of relative to the current working
        # directory.
        if (
                os.path.basename(path) == path and
                self.experiment.experiment_path is not None
        ):
            self._path = os.path.join(self.experiment.experiment_path, path)
        else:
            self._path = path
        # Open the logfile
        self.experiment.var.logfile = self._path
        if self._path not in self.experiment.data_files:
            self.experiment.data_files.append(self._path)
        self._log = safe_open(self._path, u'w')
        self._header_written = False

    def write(self, msg, newline=True):

        self._log.write(safe_decode(msg))
        if newline:
            self._log.write(u'\n')
        # Flush to avoid pending write operations
        self._log.flush()
        os.fsync(self._log)

    def write_vars(self, var_list=None):

        if var_list is None:
            var_list = self.all_vars()
        if not self._header_written:
            l = [u'"%s"' % var.replace(u'"', u'\\"') for var in var_list]
            self.write(u','.join(l))
            self._header_written = True
        l = []
        for var in var_list:
            val = self.experiment.var.get(var, _eval=False, default=u'NA')
            val = safe_decode(val)
            l.append(u'"%s"' % val.replace(u'"', u'\\"'))
        self.write(u','.join(l))


# Non PEP-8 alias for backwards compatibility
csv = Csv



# -*- coding:utf-8 -*-

"""
This file is part of OpenSesame.

OpenSesame is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenSesame is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with OpenSesame.  If not, see <http://www.gnu.org/licenses/>.
"""
from libopensesame.py3compat import *
from libopensesame.widgets._image import ImageWidget


class ImageButton(ImageWidget):

    r"""The image_button widget is a clickable image.

    __Example (OpenSesame
    script):__

    ~~~
    widget 0 0 1 1 image_button path='5.png' var='response'
    ~~~
    __Example (Python):__

    ~~~ .python
    form = Form()
    # The full path to the
    image needs to be provided.
    # self.experiment.pool can be used to retrieve
    the full path
    # to an image in the file pool.
    image_button =
    ImageButton(path=pool['5.png'], var='response')
    form.set_widget(image_button, (0,0))
    form._exec()
    ~~~

    [TOC]
    """
    def __init__(self, form, path=None, adjust=True, frame=False, image_id=None,
                 var=None):
        r"""Constructor to create a new `ImageButton` object. You do not
        generally call this constructor directly, but use the
        `ImageButton()`
        factory function, which is described here:
        [/python/common/]().

        Parameters
        ----------
        form : form
            The parent form.
        path : str, unicode, NoneType, optional
            The full path to the image. To show an image from the file pool,
            you need to first use `experiment.get_file` to determine the full
            path to the image.
        adjust : bool, optional
            Indicates whether the image should be scaled according to the size
            of the widget.
        frame : bool, optional
            Indicates whether a frame should be drawn around the widget.
        image_id : str, unicode, NoneType, optional
            An id to identify the image when it is clicked. If `None`, the path
            to the image is used as id.
        var : str, unicode, NoneType, optional
            The name of the experimental variable that should be used to log
            the widget status.
        """
        ImageWidget.__init__(self, form, path, adjust=adjust, frame=frame)
        self.image_id = path if image_id is None else image_id
        self.type = u'image_button'
        self.var = var
        self.set_var(False)

    def on_mouse_click(self, pos):
        r"""Is called whenever the user clicks on the widget. Returns the
        image_id or the path to the image if no image_id has been specified.

        Parameters
        ----------
        pos : tuple
            An (x, y) coordinate tuple.
        """
        self.theme_engine.click()
        self.set_var(True)
        return self.image_id


image_button = ImageButton
